import logging
from typing import Any, Dict, List

from wandb.errors import LaunchError

from .abstract import AbstractBuilder

__logger__ = logging.getLogger(__name__)


_WANDB_BUILDERS: List[str] = ["kaniko", "docker"]


def load_builder(builder_config: Dict[str, Any]) -> AbstractBuilder:
    builder_name = builder_config.get("type", "docker")
    if builder_name == "kaniko":
        from .kaniko import KanikoBuilder

        return KanikoBuilder(builder_config)
    elif builder_name == "docker":
        from .docker import DockerBuilder

        return DockerBuilder(builder_config)
    raise LaunchError(
        "Builder name not among available builders. Available builders: {} ".format(
            ",".join(_WANDB_BUILDERS)
        )
    )



introspection_query = '''
  query IntrospectionQuery {
    __schema {
      queryType { name }
      mutationType { name }
      subscriptionType { name }
      types {
        ...FullType
      }
      directives {
        name
        description
        locations
        args {
          ...InputValue
        }
      }
    }
  }
  fragment FullType on __Type {
    kind
    name
    description
    fields(includeDeprecated: true) {
      name
      description
      args {
        ...InputValue
      }
      type {
        ...TypeRef
      }
      isDeprecated
      deprecationReason
    }
    inputFields {
      ...InputValue
    }
    interfaces {
      ...TypeRef
    }
    enumValues(includeDeprecated: true) {
      name
      description
      isDeprecated
      deprecationReason
    }
    possibleTypes {
      ...TypeRef
    }
  }
  fragment InputValue on __InputValue {
    name
    description
    type { ...TypeRef }
    defaultValue
  }
  fragment TypeRef on __Type {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
        ofType {
          kind
          name
          ofType {
            kind
            name
            ofType {
              kind
              name
              ofType {
                kind
                name
                ofType {
                  kind
                  name
                }
              }
            }
          }
        }
      }
    }
  }
'''



"""
Internal utility for converting arguments from a launch spec or call to wandb launch
into a runnable wandb launch script
"""

import enum
import json
import logging
import os
from shlex import quote
import tempfile
from typing import Any, Dict, List, Optional

import wandb
from wandb.apis.internal import Api
import wandb.docker as docker
from wandb.errors import Error as ExecutionError, LaunchError
from wandb.sdk.lib.runid import generate_id

from . import utils


_logger = logging.getLogger(__name__)

DEFAULT_LAUNCH_METADATA_PATH = "launch_metadata.json"

# need to make user root for sagemaker, so users have access to /opt/ml directories
# that let users create artifacts and access input data
RESOURCE_UID_MAP = {"local": 1000, "sagemaker": 0}


class LaunchSource(enum.IntEnum):
    WANDB: int = 1
    GIT: int = 2
    LOCAL: int = 3
    DOCKER: int = 4


class LaunchProject:
    """A launch project specification."""

    def __init__(
        self,
        uri: Optional[str],
        api: Api,
        launch_spec: Dict[str, Any],
        target_entity: str,
        target_project: str,
        name: Optional[str],
        docker_config: Dict[str, Any],
        git_info: Dict[str, str],
        overrides: Dict[str, Any],
        resource: str,
        resource_args: Dict[str, Any],
        cuda: Optional[bool],
        run_id: Optional[str],
    ):
        if uri is not None and utils.is_bare_wandb_uri(uri):
            uri = api.settings("base_url") + uri
            _logger.info(f"Updating uri with base uri: {uri}")
        self.uri = uri
        self.api = api
        self.launch_spec = launch_spec
        self.target_entity = target_entity
        self.target_project = target_project.lower()
        self.name = name  # TODO: replace with run_id
        self.build_image: bool = docker_config.get("build_image", False)
        self.python_version: Optional[str] = docker_config.get("python_version")
        self.cuda_version: Optional[str] = docker_config.get("cuda_version")
        self._base_image: Optional[str] = docker_config.get("base_image")
        self.docker_image: Optional[str] = docker_config.get("docker_image")
        uid = RESOURCE_UID_MAP.get(resource, 1000)
        if self._base_image:
            uid = docker.get_image_uid(self._base_image)
            _logger.info(f"Retrieved base image uid {uid}")
        self.docker_user_id: int = docker_config.get("user_id", uid)
        self.git_version: Optional[str] = git_info.get("version")
        self.git_repo: Optional[str] = git_info.get("repo")
        self.override_args: Dict[str, Any] = overrides.get("args", {})
        self.override_config: Dict[str, Any] = overrides.get("run_config", {})
        self.override_artifacts: Dict[str, Any] = overrides.get("artifacts", {})
        self.override_entrypoint: Optional[EntryPoint] = None
        self.resource = resource
        self.resource_args = resource_args
        self.deps_type: Optional[str] = None
        self.cuda = cuda
        self._runtime: Optional[str] = None
        self.run_id = run_id or generate_id()
        self._entry_points: Dict[
            str, EntryPoint
        ] = {}  # todo: keep multiple entrypoint support?
        if overrides.get("entry_point") is not None:
            _logger.info("Adding override entry point")
            self.override_entrypoint = self.add_entry_point(
                overrides.get("entry_point")  # type: ignore
            )
        if self.uri is None:
            if self.docker_image is None:
                raise LaunchError("Run requires a URI or a docker image")
            self.source = LaunchSource.DOCKER
            self.project_dir = None
        elif utils._is_wandb_uri(self.uri):
            _logger.info(f"URI {self.uri} indicates a wandb uri")
            self.source = LaunchSource.WANDB
            self.project_dir = tempfile.mkdtemp()
        elif utils._is_git_uri(self.uri):
            _logger.info(f"URI {self.uri} indicates a git uri")
            self.source = LaunchSource.GIT
            self.project_dir = tempfile.mkdtemp()
        else:
            _logger.info(f"URI {self.uri} indicates a local uri")
            # assume local
            if not os.path.exists(self.uri):
                raise LaunchError(
                    "Assumed URI supplied is a local path but path is not valid"
                )
            self.source = LaunchSource.LOCAL
            self.project_dir = self.uri
        if launch_spec.get("resource_args"):
            self.resource_args = launch_spec["resource_args"]

        self.aux_dir = tempfile.mkdtemp()
        self.clear_parameter_run_config_collisions()

    @property
    def base_image(self) -> str:
        """Returns {PROJECT}_base:{PYTHON_VERSION}"""
        # TODO: this should likely be source_project when we have it...

        # don't make up a separate base image name if user provides a docker image
        if self.docker_image is not None:
            return self.docker_image

        python_version = (self.python_version or "3").replace("+", "dev")
        generated_name = "{}_base:{}".format(
            self.target_project.replace(" ", "-"), python_version
        )
        return self._base_image or generated_name

    @property
    def image_name(self) -> str:
        """Returns {PROJECT}_launch the ultimate version will
        be tagged with a sha of the git repo"""
        # TODO: this should likely be source_project when we have it...
        return f"{self.target_project}_launch"

    def clear_parameter_run_config_collisions(self) -> None:
        """Clear values from the override run config values if a matching key exists in the override arguments."""
        if not self.override_config:
            return
        keys = [key for key in self.override_config.keys()]
        for key in keys:
            if self.override_args.get(key):
                del self.override_config[key]

    def get_single_entry_point(self) -> Optional["EntryPoint"]:
        """Returns the first entrypoint for the project, or None if no entry point was provided because a docker image was provided."""
        # assuming project only has 1 entry point, pull that out
        # tmp fn until we figure out if we want to support multiple entry points or not
        if not self._entry_points:
            if not self.docker_image:
                raise LaunchError(
                    "Project must have at least one entry point unless docker image is specified."
                )
            return None
        return list(self._entry_points.values())[0]

    def add_entry_point(self, command: List[str]) -> "EntryPoint":
        """Adds an entry point to the project."""
        entry_point = command[-1]
        new_entrypoint = EntryPoint(name=entry_point, command=command)
        self._entry_points[entry_point] = new_entrypoint
        return new_entrypoint

    def _fetch_project_local(self, internal_api: Api) -> None:
        """Fetch a project (either wandb run or git repo) into a local directory, returning the path to the local project directory."""
        # these asserts are all guaranteed to pass, but are required by mypy
        assert self.source != LaunchSource.LOCAL
        assert isinstance(self.uri, str)
        assert self.project_dir is not None
        _logger.info("Fetching project locally...")
        if utils._is_wandb_uri(self.uri):
            source_entity, source_project, source_run_name = utils.parse_wandb_uri(
                self.uri
            )
            run_info = utils.fetch_wandb_project_run_info(
                source_entity, source_project, source_run_name, internal_api
            )
            program_name = run_info.get("codePath") or run_info["program"]

            if run_info.get("cudaVersion"):
                original_cuda_version = ".".join(run_info["cudaVersion"].split(".")[:2])

                if self.cuda is None:
                    # only set cuda on by default if cuda is None (unspecified), not False (user specifically requested cpu image)
                    wandb.termlog(
                        "Original wandb run {} was run with cuda version {}. Enabling cuda builds by default; to build on a CPU-only image, run again with --cuda=False".format(
                            source_run_name, original_cuda_version
                        )
                    )
                    self.cuda_version = original_cuda_version
                    self.cuda = True
                if (
                    self.cuda
                    and self.cuda_version
                    and self.cuda_version != original_cuda_version
                ):
                    wandb.termlog(
                        "Specified cuda version {} differs from original cuda version {}. Running with specified version {}".format(
                            self.cuda_version, original_cuda_version, self.cuda_version
                        )
                    )

            downloaded_code_artifact = utils.check_and_download_code_artifacts(
                source_entity,
                source_project,
                source_run_name,
                internal_api,
                self.project_dir,
            )

            if downloaded_code_artifact:
                self.build_image = True
            elif not downloaded_code_artifact:
                if not run_info["git"]:
                    raise ExecutionError(
                        "Reproducing a run requires either an associated git repo or a code artifact logged with `run.log_code()`"
                    )
                utils._fetch_git_repo(
                    self.project_dir,
                    run_info["git"]["remote"],
                    run_info["git"]["commit"],
                )
                patch = utils.fetch_project_diff(
                    source_entity, source_project, source_run_name, internal_api
                )

                if patch:
                    utils.apply_patch(patch, self.project_dir)
                # For cases where the entry point wasn't checked into git
                if not os.path.exists(os.path.join(self.project_dir, program_name)):
                    downloaded_entrypoint = utils.download_entry_point(
                        source_entity,
                        source_project,
                        source_run_name,
                        internal_api,
                        program_name,
                        self.project_dir,
                    )
                    if not downloaded_entrypoint:
                        raise LaunchError(
                            f"Entrypoint file: {program_name} does not exist, "
                            "and could not be downloaded. Please specify the entrypoint for this run."
                        )
                    # if the entrypoint is downloaded and inserted into the project dir
                    # need to rebuild image with new code
                    self.build_image = True

            if (
                "_session_history.ipynb" in os.listdir(self.project_dir)
                or ".ipynb" in program_name
            ):
                program_name = utils.convert_jupyter_notebook_to_script(
                    program_name, self.project_dir
                )

            # Download any frozen requirements
            utils.download_wandb_python_deps(
                source_entity,
                source_project,
                source_run_name,
                internal_api,
                self.project_dir,
            )

            # Specify the python runtime for jupyter2docker
            self.python_version = run_info.get("python", "3")
            if not self._entry_points:
                _, ext = os.path.splitext(program_name)
                if ext == ".py":
                    entry_point = ["python", program_name]
                elif ext == ".sh":
                    command = os.environ.get("SHELL", "bash")
                    entry_point = [command, program_name]
                else:
                    raise LaunchError(f"Unsupported entrypoint: {program_name}")
                self.add_entry_point(entry_point)
            self.override_args = utils.merge_parameters(
                self.override_args, run_info["args"]
            )
        else:
            assert utils._GIT_URI_REGEX.match(self.uri), (
                "Non-wandb URI %s should be a Git URI" % self.uri
            )
            if not self._entry_points:
                wandb.termlog(
                    "Entry point for repo not specified, defaulting to python main.py"
                )
                self.add_entry_point(["python", "main.py"])
            utils._fetch_git_repo(self.project_dir, self.uri, self.git_version)


class EntryPoint:
    """An entry point into a wandb launch specification."""

    def __init__(self, name: str, command: List[str]):
        self.name = name
        self.command = command

    def compute_command(self, user_parameters: Optional[Dict[str, Any]]) -> List[str]:
        """Converts user parameter dictionary to a string."""
        command_arr = []
        command_arr += self.command
        extras = compute_command_args(user_parameters)
        command_arr += extras
        return command_arr


def compute_command_args(parameters: Optional[Dict[str, Any]]) -> List[str]:
    arr: List[str] = []
    if parameters is None:
        return arr
    for key, value in parameters.items():
        if value is not None:
            arr.append(f"--{key}")
            arr.append(quote(str(value)))
        else:
            arr.append(f"--{key}")
    return arr


def get_entry_point_command(
    entry_point: Optional["EntryPoint"], parameters: Dict[str, Any]
) -> List[str]:
    """Returns the shell command to execute in order to run the specified entry point.

    Arguments:
    entry_point: Entry point to run
    parameters: Parameters (dictionary) for the entry point command

    Returns:
        List of strings representing the shell command to be executed
    """
    if entry_point is None:
        return []
    return entry_point.compute_command(parameters)


def create_project_from_spec(launch_spec: Dict[str, Any], api: Api) -> LaunchProject:
    """Constructs a LaunchProject instance using a launch spec.

    Arguments:
    launch_spec: Dictionary representation of launch spec
    api: Instance of wandb.apis.internal Api

    Returns:
        An initialized `LaunchProject` object
    """

    name: Optional[str] = None
    if launch_spec.get("name"):
        name = launch_spec["name"]
    return LaunchProject(
        launch_spec.get("uri"),
        api,
        launch_spec,
        launch_spec["entity"],
        launch_spec["project"],
        name,
        launch_spec.get("docker", {}),
        launch_spec.get("git", {}),
        launch_spec.get("overrides", {}),
        launch_spec.get("resource", "local"),
        launch_spec.get("resource_args", {}),
        launch_spec.get("cuda", None),
        launch_spec.get("run_id", None),
    )


def fetch_and_validate_project(
    launch_project: LaunchProject, api: Api
) -> LaunchProject:
    """Fetches a project into a local directory, adds the config values to the directory, and validates the first entrypoint for the project.

    Arguments:
    launch_project: LaunchProject to fetch and validate.
    api: Instance of wandb.apis.internal Api

    Returns:
        A validated `LaunchProject` object.

    """
    if launch_project.source == LaunchSource.DOCKER:
        return launch_project
    if launch_project.source == LaunchSource.LOCAL:
        if not launch_project._entry_points:
            wandb.termlog(
                "Entry point for local directory not specified, defaulting to `python main.py`"
            )
            launch_project.add_entry_point(["python", "main.py"])
    else:
        launch_project._fetch_project_local(internal_api=api)

    assert launch_project.project_dir is not None
    # this prioritizes pip, and we don't support any cases where both are present
    # conda projects when uploaded to wandb become pip projects via requirements.frozen.txt, wandb doesn't preserve conda envs
    if os.path.exists(
        os.path.join(launch_project.project_dir, "requirements.txt")
    ) or os.path.exists(
        os.path.join(launch_project.project_dir, "requirements.frozen.txt")
    ):
        launch_project.deps_type = "pip"
    elif os.path.exists(os.path.join(launch_project.project_dir, "environment.yml")):
        launch_project.deps_type = "conda"

    return launch_project


def create_metadata_file(
    launch_project: LaunchProject,
    image_uri: str,
    sanitized_entrypoint_str: str,
    docker_args: Dict[str, Any],
    sanitized_dockerfile_contents: str,
) -> None:
    assert launch_project.project_dir is not None
    with open(
        os.path.join(launch_project.project_dir, DEFAULT_LAUNCH_METADATA_PATH),
        "w",
    ) as f:
        json.dump(
            {
                **launch_project.launch_spec,
                "image_uri": image_uri,
                "command": sanitized_entrypoint_str,
                "docker_args": docker_args,
                "dockerfile_contents": sanitized_dockerfile_contents,
            },
            f,
        )



import re

import wandb


def monitor():
    vcr = wandb.util.get_module(
        "gym.wrappers.monitoring.video_recorder",
        required="Couldn't import the gym python package, install with pip install gym",
    )
    vcr.ImageEncoder.orig_close = vcr.ImageEncoder.close

    def close(self):
        vcr.ImageEncoder.orig_close(self)
        m = re.match(r".+(video\.\d+).+", self.output_path)
        if m:
            key = m.group(1)
        else:
            key = "videos"
        wandb.log({key: wandb.Video(self.output_path)})

    vcr.ImageEncoder.close = close
    wandb.patched["gym"].append(
        ["gym.wrappers.monitoring.video_recorder.ImageEncoder", "close"]
    )



import wandb
from wandb import util
from wandb.plots.utils import (
    test_missing,
    test_types,
    encode_labels,
    deprecation_notice,
)


def named_entity(docs):
    """
    Adds support for spaCy's entity visualizer, which highlights named
        entities and their labels in a text.

    Arguments:
     docs (list, Doc, Span): Document(s) to visualize.

    Returns:
     Nothing. To see plots, go to your W&B run page.

    Example:
     wandb.log({'NER': wandb.plots.NER(docs=doc)})
    """
    deprecation_notice()

    spacy = util.get_module(
        "spacy",
        required="part_of_speech requires the spacy library, install with `pip install spacy`",
    )
    en_core_web_md = util.get_module(
        "en_core_web_md",
        required="part_of_speech requires the en_core_web_md library, install with `python -m spacy download en_core_web_md`",
    )
    nlp = en_core_web_md.load()

    if test_missing(docs=docs):
        # and test_types(docs=docs)):
        wandb.termlog("Visualizing named entity recognition.")
        html = spacy.displacy.render(
            nlp(str(docs)), style="ent", page=True, minify=True
        )
        return wandb.Html(html)



from copy import deepcopy
import inspect
import json
import re
from typing import Any, Dict, List as LList
from typing import Optional, Union
import urllib

import wandb
from wandb.sdk.lib import ipython

from .mutations import CREATE_PROJECT, UPSERT_VIEW
from .util import (
    Attr,
    Base,
    Block,
    coalesce,
    fix_collisions,
    generate_name,
    nested_get,
    nested_set,
    Panel,
)
from .validators import (
    AGGFUNCS,
    Between,
    CODE_COMPARE_DIFF,
    FONT_SIZES,
    LEGEND_POSITIONS,
    Length,
    LINEPLOT_STYLES,
    MARKS,
    OneOf,
    RANGEFUNCS,
    SMOOTHING_TYPES,
    TypeValidator,
)


class LineKey:
    def __init__(self, key: str) -> None:
        self.key = key

    def __hash__(self) -> int:
        return hash(self.key)

    def __repr__(self) -> str:
        return f'LineKey(key="{self.key}")'

    @classmethod
    def from_run(cls, run: "wandb.apis.public.Run", metric: str) -> "LineKey":
        key = f"{run.id}:{metric}"
        return cls(key)

    @classmethod
    def from_panel_agg(cls, runset: "RunSet", panel: "Panel", metric: str) -> "LineKey":
        key = f"{runset.id}-config:group:{panel.groupby}:null:{metric}"
        return cls(key)

    @classmethod
    def from_runset_agg(cls, runset: "RunSet", metric: str) -> "LineKey":
        groupby = runset.groupby
        if runset.groupby is None:
            groupby = "null"

        key = f"{runset.id}-run:group:{groupby}:{metric}"
        return cls(key)


class RGBA(Base):
    r: int = Attr(validators=[Between(0, 255)])
    g: int = Attr(validators=[Between(0, 255)])
    b: int = Attr(validators=[Between(0, 255)])
    a: Union[int, float] = Attr(validators=[Between(0, 1)])

    def __init__(
        self, r: int, g: int, b: int, a: Union[int, float] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "RGBA":
        color = d.get("transparentColor").replace(" ", "")
        r, g, b, a = re.split(r"\(|\)|,", color)[1:-1]
        r, g, b, a = int(r), int(g), int(b), float(a)
        return cls(r, g, b, a)

    @property
    def spec(self) -> Dict[str, Any]:
        return {
            "color": f"rgb({self.r}, {self.g}, {self.b})",
            "transparentColor": f"rgba({self.r}, {self.g}, {self.b}, {self.a})",
        }


class UnknownPanel(Panel):
    @property
    def view_type(self) -> str:
        return "UNKNOWN PANEL"


class LinePlot(Panel):
    def __init__(
        self,
        title: Optional[str] = None,
        x: Optional[str] = None,
        y: Optional[list] = None,
        range_x: Union[list, tuple] = (None, None),
        range_y: Union[list, tuple] = (None, None),
        log_x: Optional[bool] = None,
        log_y: Optional[bool] = None,
        title_x: Optional[str] = None,
        title_y: Optional[str] = None,
        ignore_outliers: Optional[bool] = None,
        groupby: Optional[str] = None,
        groupby_aggfunc: Optional[str] = None,
        groupby_rangefunc: Optional[str] = None,
        smoothing_factor: Optional[float] = None,
        smoothing_type: Optional[str] = None,
        smoothing_show_original: Optional[bool] = None,
        max_runs_to_show: Optional[int] = None,
        custom_expressions: Optional[str] = None,
        plot_type: Optional[str] = None,
        font_size: Optional[str] = None,
        legend_position: Optional[str] = None,
        legend_template: Optional[str] = None,
        aggregate: Optional[bool] = None,
        xaxis_expression: Optional[str] = None,
        line_titles: Optional[dict] = None,
        line_marks: Optional[dict] = None,
        line_colors: Optional[dict] = None,
        line_widths: Optional[dict] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.title = title
        self.x = x
        self.y = y
        self.range_x = range_x
        self.range_y = range_y
        self.log_x = log_x
        self.log_y = log_y
        self.title_x = title_x
        self.title_y = title_y
        self.ignore_outliers = ignore_outliers
        self.groupby = groupby
        self.groupby_aggfunc = groupby_aggfunc
        self.groupby_rangefunc = groupby_rangefunc
        self.smoothing_factor = smoothing_factor
        self.smoothing_type = smoothing_type
        self.smoothing_show_original = smoothing_show_original
        self.max_runs_to_show = max_runs_to_show
        self.custom_expressions = custom_expressions
        self.plot_type = plot_type
        self.font_size = font_size
        self.legend_position = legend_position
        self.legend_template = legend_template
        self.aggregate = aggregate
        self.xaxis_expression = xaxis_expression
        self.line_titles = line_titles
        self.line_marks = line_marks
        self.line_colors = line_colors
        self.line_widths = line_widths

    title: Optional[str] = Attr(
        json_path="spec.config.chartTitle",
    )
    x: Optional[str] = Attr(json_path="spec.config.xAxis")
    y: Optional[list] = Attr(json_path="spec.config.metrics")
    range_x: Union[list, tuple] = Attr(
        json_path=["spec.config.xAxisMin", "spec.config.xAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    range_y: Union[list, tuple] = Attr(
        json_path=["spec.config.yAxisMin", "spec.config.yAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    log_x: Optional[bool] = Attr(json_path="spec.config.xLogScale")
    log_y: Optional[bool] = Attr(json_path="spec.config.xLogScale")
    title_x: Optional[str] = Attr(json_path="spec.config.xAxisTitle")
    title_y: Optional[str] = Attr(json_path="spec.config.yAxisTitle")
    ignore_outliers: Optional[bool] = Attr(json_path="spec.config.ignoreOutliers")
    groupby: Optional[str] = Attr(json_path="spec.config.groupBy")
    groupby_aggfunc: Optional[str] = Attr(
        json_path="spec.config.groupAgg",
        validators=[OneOf(AGGFUNCS)],
    )
    groupby_rangefunc: Optional[str] = Attr(
        json_path="spec.config.groupArea",
        validators=[OneOf(RANGEFUNCS)],
    )
    smoothing_factor: Optional[float] = Attr(json_path="spec.config.smoothingWeight")
    smoothing_type: Optional[str] = Attr(
        json_path="spec.config.smoothingType",
        validators=[OneOf(SMOOTHING_TYPES)],
    )
    smoothing_show_original: Optional[bool] = Attr(
        json_path="spec.config.showOriginalAfterSmoothing"
    )
    max_runs_to_show: Optional[int] = Attr(json_path="spec.config.smoothingType")
    custom_expressions: Optional[str] = Attr(json_path="spec.config.expressions")
    plot_type: Optional[str] = Attr(
        json_path="spec.config.plotType",
        validators=[OneOf(LINEPLOT_STYLES)],
    )
    font_size: Optional[str] = Attr(
        json_path="spec.config.fontSize",
        validators=[OneOf(FONT_SIZES)],
    )
    legend_position: Optional[str] = Attr(
        json_path="spec.config.legendPosition",
        validators=[OneOf(LEGEND_POSITIONS)],
    )
    legend_template: Optional[str] = Attr(json_path="spec.config.legendTemplate")
    # Attr( json_path="spec.config.startingXAxis")
    # Attr( json_path="spec.config.useLocalSmoothing")
    # Attr( json_path="spec.config.useGlobalSmoothingWeight")
    # Attr( json_path="spec.config.legendFields")
    aggregate: Optional[bool] = Attr(json_path="spec.config.aggregate")
    # Attr( json_path="spec.config.aggregateMetrics")
    # Attr( json_path="spec.config.metricRegex")
    # Attr( json_path="spec.config.useMetricRegex")
    # Attr( json_path="spec.config.yAxisAutoRange")
    # Attr( json_path="spec.config.groupRunsLimit")
    xaxis_expression: Optional[str] = Attr(json_path="spec.config.xExpression")
    # Attr( json_path="spec.config.colorEachMetricDifferently")
    # Attr( json_path="spec.config.showLegend")

    line_titles: Optional[dict] = Attr(
        json_path="spec.config.overrideSeriesTitles",
        validators=[
            TypeValidator(LineKey, how="keys"),
            TypeValidator(str, how="values"),
        ],
    )
    line_marks: Optional[dict] = Attr(
        json_path="spec.config.overrideMarks",
        validators=[
            TypeValidator(LineKey, how="keys"),
            OneOf(MARKS, how="values"),
        ],
    )
    line_colors: Optional[dict] = Attr(
        json_path="spec.config.overrideColors",
        validators=[
            TypeValidator(LineKey, how="keys"),
            TypeValidator(RGBA, how="values"),
        ],
    )
    line_widths: Optional[dict] = Attr(
        json_path="spec.config.overrideLineWidths",
        validators=[
            TypeValidator(LineKey, how="keys"),
            TypeValidator(Union[int, float], how="values"),
            Between(0.5, 3.0, how="values"),
        ],
    )

    @x.getter
    def x(self):
        json_path = self._get_path("x")
        value = nested_get(self, json_path)
        return self.panel_metrics_helper.back_to_front(value)

    @x.setter
    def x(self, value):
        json_path = self._get_path("x")
        value = self.panel_metrics_helper.front_to_back(value)
        nested_set(self, json_path, value)

    @y.getter
    def y(self):
        json_path = self._get_path("y")
        value = nested_get(self, json_path)
        if value is None:
            return value
        return [self.panel_metrics_helper.back_to_front(v) for v in value]

    @y.setter
    def y(self, value):
        json_path = self._get_path("y")
        if value is not None:
            value = [self.panel_metrics_helper.front_to_back(v) for v in value]
        nested_set(self, json_path, value)

    @property
    def view_type(self):
        return "Run History Line Plot"


class ScatterPlot(Panel):
    def __init__(
        self,
        title=None,
        x=None,
        y=None,
        z=None,
        range_x=(None, None),
        range_y=(None, None),
        range_z=(None, None),
        log_x=None,
        log_y=None,
        log_z=None,
        running_ymin=None,
        running_ymax=None,
        running_ymean=None,
        legend_template=None,
        gradient=None,
        font_size=None,
        regression=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.title = title
        self.x = x
        self.y = y
        self.z = z
        self.range_x = range_x
        self.range_y = range_y
        self.range_z = range_z
        self.log_x = log_x
        self.log_y = log_y
        self.log_z = log_z
        self.running_ymin = running_ymin
        self.running_ymax = running_ymax
        self.running_ymean = running_ymean
        self.legend_template = legend_template
        self.gradient = gradient
        self.font_size = font_size
        self.regression = regression

    title: Optional[str] = Attr(json_path="spec.config.chartTitle")
    x: Optional[str] = Attr(json_path="spec.config.xAxis")
    y: Optional[str] = Attr(json_path="spec.config.yAxis")
    z: Optional[str] = Attr(json_path="spec.config.zAxis")
    range_x: Union[list, tuple] = Attr(
        json_path=["spec.config.xAxisMin", "spec.config.xAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    range_y: Union[list, tuple] = Attr(
        json_path=["spec.config.yAxisMin", "spec.config.yAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    range_z: Union[list, tuple] = Attr(
        json_path=["spec.config.zAxisMin", "spec.config.zAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    log_x: Optional[bool] = Attr(json_path="spec.config.xAxisLogScale")
    log_y: Optional[bool] = Attr(json_path="spec.config.yAxisLogScale")
    log_z: Optional[bool] = Attr(json_path="spec.config.zAxisLogScale")
    running_ymin: Optional[bool] = Attr(json_path="spec.config.showMaxYAxisLine")
    running_ymax: Optional[bool] = Attr(json_path="spec.config.showMinYAxisLine")
    running_ymean: Optional[bool] = Attr(json_path="spec.config.showAvgYAxisLine")
    legend_template: Optional[str] = Attr(json_path="spec.config.legendTemplate")
    gradient: Optional[dict] = Attr(
        json_path="spec.config.customGradient",
        validators=[TypeValidator(RGBA, how="values")],
    )
    # color: ... = Attr(json_path="spec.config.color")
    # range_color: ... = Attr(
    #     ["spec.config.minColor", "spec.config.maxColor"],
    #     (list, tuple),
    #     validators=[Length(2), TypeValidator((int, float), how='keys')],
    # )

    # Attr(json_path="spec.config.legendFields")
    font_size: Optional[str] = Attr(
        json_path="spec.config.fontSize",
        validators=[OneOf(FONT_SIZES)],
    )
    # Attr(json_path="spec.config.yAxisLineSmoothingWeight")
    regression: Optional[bool] = Attr(json_path="spec.config.showLinearRegression")

    @x.getter
    def x(self):
        json_path = self._get_path("x")
        value = nested_get(self, json_path)
        return self.panel_metrics_helper.special_back_to_front(value)

    @x.setter
    def x(self, value):
        json_path = self._get_path("x")
        value = self.panel_metrics_helper.special_front_to_back(value)
        nested_set(self, json_path, value)

    @y.getter
    def y(self):
        json_path = self._get_path("y")
        value = nested_get(self, json_path)
        return self.panel_metrics_helper.special_back_to_front(value)

    @y.setter
    def y(self, value):
        json_path = self._get_path("y")
        value = self.panel_metrics_helper.special_front_to_back(value)
        nested_set(self, json_path, value)

    @z.getter
    def z(self):
        json_path = self._get_path("z")
        value = nested_get(self, json_path)
        return self.panel_metrics_helper.special_back_to_front(value)

    @z.setter
    def z(self, value):
        json_path = self._get_path("z")
        value = self.panel_metrics_helper.special_front_to_back(value)
        nested_set(self, json_path, value)

    @property
    def view_type(self) -> str:
        return "Scatter Plot"


class BarPlot(Panel):
    def __init__(
        self,
        title=None,
        metrics=None,
        vertical=None,
        range_x=(None, None),
        title_x=None,
        title_y=None,
        groupby=None,
        groupby_aggfunc=None,
        groupby_rangefunc=None,
        max_runs_to_show=None,
        max_bars_to_show=None,
        custom_expressions=None,
        legend_template=None,
        font_size=None,
        line_titles=None,
        line_colors=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.title = title
        self.metrics = metrics
        self.vertical = vertical
        self.range_x = range_x
        self.title_x = title_x
        self.title_y = title_y
        self.groupby = groupby
        self.groupby_aggfunc = groupby_aggfunc
        self.groupby_rangefunc = groupby_rangefunc
        self.max_runs_to_show = max_runs_to_show
        self.max_bars_to_show = max_bars_to_show
        self.custom_expressions = custom_expressions
        self.legend_template = legend_template
        self.font_size = font_size
        self.line_titles = line_titles
        self.line_colors = line_colors

    title: Optional[str] = Attr(json_path="spec.config.chartTitle")
    metrics: Optional[list] = Attr(
        json_path="spec.config.metrics",
        validators=[TypeValidator(str, how="keys")],
    )
    vertical: Optional[bool] = Attr(json_path="spec.config.vertical")
    range_x: Union[list, tuple] = Attr(
        json_path=["spec.config.xAxisMin", "spec.config.xAxisMax"],
        validators=[
            Length(2),
            TypeValidator(Optional[Union[int, float]], how="keys"),
        ],
    )
    title_x: Optional[str] = Attr(json_path="spec.config.xAxisTitle")
    title_y: Optional[str] = Attr(json_path="spec.config.yAxisTitle")
    groupby: Optional[str] = Attr(json_path="spec.config.groupBy")
    groupby_aggfunc: Optional[str] = Attr(
        json_path="spec.config.groupAgg",
        validators=[OneOf(AGGFUNCS)],
    )
    groupby_rangefunc: Optional[str] = Attr(
        json_path="spec.config.groupArea",
        validators=[OneOf(RANGEFUNCS)],
    )
    max_runs_to_show: Optional[int] = Attr(json_path="spec.config.limit")
    max_bars_to_show: Optional[int] = Attr(json_path="spec.config.barLimit")
    custom_expressions: Optional[str] = Attr(json_path="spec.config.expressions")
    legend_template: Optional[str] = Attr(json_path="spec.config.legendTemplate")
    font_size: Optional[str] = Attr(
        json_path="spec.config.fontSize",
        validators=[OneOf(FONT_SIZES)],
    )
    # Attr(json_path="spec.config.limit")
    # Attr(json_path="spec.config.barLimit")
    # Attr(json_path="spec.config.aggregate")
    # Attr(json_path="spec.config.aggregateMetrics")
    # Attr(json_path="spec.config.groupRunsLimit")
    # Attr(json_path="spec.config.plotStyle")
    # Attr(json_path="spec.config.legendFields")
    # Attr(json_path="spec.config.colorEachMetricDifferently")

    line_titles: Optional[dict] = Attr(
        json_path="spec.config.overrideSeriesTitles",
        validators=[
            TypeValidator(LineKey, how="keys"),
            TypeValidator(str, how="values"),
        ],
    )
    line_colors: Optional[dict] = Attr(
        json_path="spec.config.overrideColors",
        validators=[
            TypeValidator(LineKey, how="keys"),
            TypeValidator(RGBA, how="values"),
        ],
    )

    @metrics.getter
    def metrics(self):
        json_path = self._get_path("metrics")
        value = nested_get(self, json_path)
        if value is None:
            return value
        return [self.panel_metrics_helper.back_to_front(v) for v in value]

    @metrics.setter
    def metrics(self, value):
        json_path = self._get_path("metrics")
        if value is not None:
            value = [self.panel_metrics_helper.front_to_back(v) for v in value]
        nested_set(self, json_path, value)

    @property
    def view_type(self) -> str:
        return "Bar Chart"


class ScalarChart(Panel):
    def __init__(
        self,
        title=None,
        metric=None,
        groupby_aggfunc=None,
        groupby_rangefunc=None,
        custom_expressions=None,
        legend_template=None,
        font_size=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.title = title
        self.metric = coalesce(metric, "")
        self.groupby_aggfunc = groupby_aggfunc
        self.groupby_rangefunc = groupby_rangefunc
        self.custom_expressions = custom_expressions
        self.legend_template = legend_template
        self.font_size = font_size

    title: Optional[str] = Attr(json_path="spec.config.chartTitle")
    metric: str = Attr(json_path="spec.config.metrics")
    groupby_aggfunc: Optional[str] = Attr(
        json_path="spec.config.groupAgg",
        validators=[OneOf(AGGFUNCS)],
    )
    groupby_rangefunc: Optional[str] = Attr(
        json_path="spec.config.groupArea",
        validators=[OneOf(RANGEFUNCS)],
    )
    custom_expressions: Optional[str] = Attr(json_path="spec.config.expressions")
    legend_template: Optional[str] = Attr(json_path="spec.config.legendTemplate")

    # Attr(json_path="spec.config.aggregate")
    # Attr(json_path="spec.config.aggregateMetrics")
    # Attr(json_path="spec.config.groupBy")
    # Attr(json_path="spec.config.groupRunsLimit")
    # Attr(json_path="spec.config.legendFields")
    # Attr(json_path="spec.config.showLegend")
    font_size: Optional[str] = Attr(
        json_path="spec.config.fontSize",
        validators=[OneOf(FONT_SIZES)],
    )

    @metric.getter
    def metric(self):
        json_path = self._get_path("metric")
        value = nested_get(self, json_path)[0]
        return self.panel_metrics_helper.back_to_front(value)

    @metric.setter
    def metric(self, new_metrics):
        json_path = self._get_path("metric")
        new_metrics = self.panel_metrics_helper.front_to_back(new_metrics)
        nested_set(self, json_path, [new_metrics])

    @property
    def view_type(self) -> str:
        return "Scalar Chart"


class CodeComparer(Panel):
    def __init__(self, diff=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.diff = diff

    diff: Optional[str] = Attr(
        json_path="spec.config.diff",
        validators=[OneOf(CODE_COMPARE_DIFF)],
    )

    @property
    def view_type(self) -> str:
        return "Code Comparer"


class PCColumn(Base):
    def __init__(
        self, metric, name=None, ascending=None, log_scale=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.metric = metric
        self.name = name
        self.ascending = ascending
        self.log_scale = log_scale

    metric: str = Attr(json_path="spec.accessor")
    name: Optional[str] = Attr(json_path="spec.displayName")
    ascending: Optional[bool] = Attr(json_path="spec.inverted")
    log_scale: Optional[bool] = Attr(json_path="spec.log")

    @classmethod
    def from_json(cls, spec):
        obj = cls(metric=spec["accessor"])
        obj._spec = spec
        return obj


class ParallelCoordinatesPlot(Panel):
    def __init__(self, columns=None, title=None, font_size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.columns = coalesce(columns, [])
        self.title = title
        self.font_size = font_size

    columns: list = Attr(
        json_path="spec.config.columns",
        validators=[TypeValidator(PCColumn, how="keys")],
    )
    title: Optional[str] = Attr(json_path="spec.config.chartTitle")

    # Attr(json_path="spec.config.dimensions")
    # Attr(json_path="spec.config.customGradient")
    # Attr(json_path="spec.config.gradientColor")
    # Attr(json_path="spec.config.legendFields")
    font_size: Optional[str] = Attr(
        json_path="spec.config.fontSize",
        validators=[OneOf(FONT_SIZES)],
    )

    @columns.getter
    def columns(self):
        json_path = self._get_path("columns")
        specs = nested_get(self, json_path)
        return [PCColumn.from_json(cspec) for cspec in specs]

    @columns.setter
    def columns(self, new_columns):
        json_path = self._get_path("columns")
        specs = [c.spec for c in new_columns]
        nested_set(self, json_path, specs)

    @property
    def view_type(self) -> str:
        return "Parallel Coordinates Plot"


class ParameterImportancePlot(Panel):
    def __init__(self, with_respect_to=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_respect_to = coalesce(with_respect_to, "Created Timestamp")

    with_respect_to: str = Attr(json_path="spec.config.targetKey")

    @with_respect_to.getter
    def with_respect_to(self):
        json_path = self._get_path("with_respect_to")
        value = nested_get(self, json_path)
        return self.panel_metrics_helper.back_to_front(value)

    @with_respect_to.setter
    def with_respect_to(self, value):
        json_path = self._get_path("with_respect_to")
        value = self.panel_metrics_helper.front_to_back(value)
        nested_set(self, json_path, value)

    @property
    def view_type(self) -> str:
        return "Parameter Importance"


class RunComparer(Panel):
    def __init__(self, diff_only=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.diff_only = diff_only

    diff_only: Optional[str] = Attr(
        json_path="spec.config.diffOnly",
        validators=[OneOf(["split", None])],
    )

    @property
    def view_type(self) -> str:
        return "Run Comparer"


class MediaBrowser(Panel):
    def __init__(self, num_columns=None, media_keys=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_columns = num_columns
        self.media_keys = media_keys

    num_columns: Optional[int] = Attr(json_path="spec.config.columnCount")
    media_keys: Optional[str] = Attr(json_path="spec.config.media_keys")

    # Attr(json_path="spec.config.chartTitle")
    # Attr(json_path="spec.config.stepIndex")
    # Attr(json_path="spec.config.mediaIndex")
    # Attr(json_path="spec.config.actualSize")
    # Attr(json_path="spec.config.fitToDimension")
    # Attr(json_path="spec.config.pixelated")
    # Attr(json_path="spec.config.mode")
    # Attr(json_path="spec.config.gallerySettings")
    # Attr(json_path="spec.config.gridSettings")
    # Attr(json_path="spec.config.selection")
    # Attr(json_path="spec.config.page")
    # Attr(json_path="spec.config.tileLayout")
    # Attr(json_path="spec.config.stepStrideLength")
    # Attr(json_path="spec.config.snapToExistingStep")
    # Attr(json_path="spec.config.maxGalleryItems")
    # Attr(json_path="spec.config.maxYAxisCount")
    # Attr(json_path="spec.config.moleculeConfig")
    # Attr(json_path="spec.config.segmentationMaskConfig")
    # Attr(json_path="spec.config.boundingBoxConfig")

    @property
    def view_type(self) -> str:
        return "Media Browser"


class MarkdownPanel(Panel):
    def __init__(self, markdown=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.markdown = markdown

    markdown: Optional[str] = Attr(json_path="spec.config.value")

    @property
    def view_type(self) -> str:
        return "Markdown Panel"


class ConfusionMatrix(Panel):
    @property
    def view_type(self) -> str:
        return "Confusion Matrix"


class DataFrames(Panel):
    @property
    def view_type(self) -> str:
        return "Data Frame Table"


class MultiRunTable(Panel):
    @property
    def view_type(self) -> str:
        return "Multi Run Table"


class Vega(Panel):
    @property
    def view_type(self) -> str:
        return "Vega"


class Vega2(Panel):
    @property
    def view_type(self) -> str:
        return "Vega2"


class Vega3(Panel):
    @property
    def view_type(self) -> str:
        return "Vega3"


class WeavePanel(Panel):
    @property
    def view_type(self) -> str:
        return "Weave"


class RunSet(Base):
    def __init__(
        self,
        entity=None,
        project="",
        name="Run set",
        query="",
        filters=None,
        groupby=None,
        order=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._spec = self._default_runset_spec()
        self.query_generator = wandb.apis.public.QueryGenerator()
        self.pm_query_generator = wandb.apis.public.PythonMongoishQueryGenerator(self)

        # self.entity = entity if entity != "" else wandb.Api().default_entity
        self.entity = coalesce(entity, wandb.Api().default_entity, "")
        self.project = project
        self.name = name
        self.query = query
        self.filters = coalesce(filters, self._default_filters())
        self.groupby = coalesce(groupby, self._default_groupby())
        self.order = coalesce(order, self._default_order())

    entity: str = Attr(json_path="spec.project.entityName")
    project: str = Attr(json_path="spec.project.name")
    name: str = Attr(json_path="spec.name")
    query: str = Attr(json_path="spec.search.query")
    filters: dict = Attr(json_path="spec.filters")
    groupby: list = Attr(json_path="spec.grouping")
    order: list = Attr(json_path="spec.sort")

    @filters.getter
    def filters(self):
        json_path = self._get_path("filters")
        filter_specs = nested_get(self, json_path)
        return self.query_generator.filter_to_mongo(filter_specs)

    @filters.setter
    def filters(self, new_filters):
        json_path = self._get_path("filters")
        new_filter_specs = self.query_generator.mongo_to_filter(new_filters)
        nested_set(self, json_path, new_filter_specs)

    def set_filters_with_python_expr(self, expr):
        self.filters = self.pm_query_generator.python_to_mongo(expr)
        return self

    @groupby.getter
    def groupby(self):
        json_path = self._get_path("groupby")
        groupby_specs = nested_get(self, json_path)
        cols = [self.query_generator.key_to_server_path(k) for k in groupby_specs]
        return [self.pm_query_generator.back_to_front(c) for c in cols]

    @groupby.setter
    def groupby(self, new_groupby):
        json_path = self._get_path("groupby")
        cols = [self.pm_query_generator.front_to_back(g) for g in new_groupby]
        new_groupby_specs = [self.query_generator.server_path_to_key(c) for c in cols]
        nested_set(self, json_path, new_groupby_specs)

    @order.getter
    def order(self):
        json_path = self._get_path("order")
        order_specs = nested_get(self, json_path)
        cols = self.query_generator.keys_to_order(order_specs)
        return [c[0] + self.pm_query_generator.back_to_front(c[1:]) for c in cols]

    @order.setter
    def order(self, new_orders):
        json_path = self._get_path("order")
        cols = [o[0] + self.pm_query_generator.front_to_back(o[1:]) for o in new_orders]
        new_order_specs = self.query_generator.order_to_keys(cols)
        nested_set(self, json_path, new_order_specs)

    @property
    def _runs_config(self) -> dict:
        # breakpoint()
        return {k: v for run in self.runs for k, v in run.config.items()}

    @property
    def runs(self) -> wandb.apis.public.Runs:
        return wandb.apis.public.Runs(wandb.Api().client, self.entity, self.project)

    @staticmethod
    def _default_filters():
        return {"$or": [{"$and": []}]}

    @staticmethod
    def _default_groupby():
        return []

    @staticmethod
    def _default_order():
        return ["-CreatedTimestamp"]

    @staticmethod
    def _default_runset_spec():
        return {
            "runFeed": {
                "version": 2,
                "columnVisible": {"run:name": False},
                "columnPinned": {},
                "columnWidths": {},
                "columnOrder": [],
                "pageSize": 10,
                "onlyShowSelected": False,
            },
            "enabled": True,
            "selections": {"root": 1, "bounds": [], "tree": []},
            "expandedRowAddresses": [],
        }


class UnknownBlock(Block):
    pass


class PanelGrid(Block):
    def __init__(self, runsets=None, panels=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._spec = self._default_panel_grid_spec()
        self.runsets = coalesce(runsets, self._default_runsets())
        self.panels = coalesce(panels, self._default_panels())

    runsets: list = Attr(json_path="spec.metadata.runSets")
    panels: list = Attr(json_path="spec.metadata.panelBankSectionConfig.panels")

    @runsets.getter
    def runsets(self):
        json_path = self._get_path("runsets")
        specs = nested_get(self, json_path)
        return [RunSet.from_json(spec) for spec in specs]

    @runsets.setter
    def runsets(self, new_runsets):
        json_path = self._get_path("runsets")
        new_specs = [rs.spec for rs in new_runsets]
        nested_set(self, json_path, new_specs)

    @panels.getter
    def panels(self):
        json_path = self._get_path("panels")
        specs = nested_get(self, json_path)
        panels = []
        for pspec in specs:
            cls = panel_mapping.get(pspec["viewType"], UnknownPanel)
            if cls is UnknownPanel:
                wandb.termwarn(
                    inspect.cleandoc(
                        f"""
                        UNKNOWN PANEL DETECTED
                            This can happen if we have added new panels, but you are using an older version of the SDK.
                            If your report is loading normally, you can safely ignore this message (but we recommend not touching UnknownPanel)
                            If you think this is an error, please file a bug report including your SDK version ({wandb.__version__}) and this spec ({pspec})
                        """
                    )
                )
            panels.append(cls.from_json(pspec))
        return panels

    @panels.setter
    def panels(self, new_panels):
        json_path = self._get_path("panels")

        # For PC and Scatter, we need to use slightly different values, so update if possible.
        # This only happens on set, and only when assigned to a panel grid because that is
        # the earliest time that we have a runset to check what kind of metric is being assigned.
        new_panels = self._get_specific_keys_for_certain_plots(new_panels, setting=True)

        new_specs = [p.spec for p in fix_collisions(new_panels)]
        nested_set(self, json_path, new_specs)

    @staticmethod
    def _default_panel_grid_spec():
        return {
            "type": "panel-grid",
            "children": [{"text": ""}],
            "metadata": {
                "openViz": True,
                "panels": {
                    "views": {"0": {"name": "Panels", "defaults": [], "config": []}},
                    "tabs": ["0"],
                },
                "panelBankConfig": {
                    "state": 0,
                    "settings": {
                        "autoOrganizePrefix": 2,
                        "showEmptySections": False,
                        "sortAlphabetically": False,
                    },
                    "sections": [
                        {
                            "name": "Hidden Panels",
                            "isOpen": False,
                            "panels": [],
                            "type": "flow",
                            "flowConfig": {
                                "snapToColumns": True,
                                "columnsPerPage": 3,
                                "rowsPerPage": 2,
                                "gutterWidth": 16,
                                "boxWidth": 460,
                                "boxHeight": 300,
                            },
                            "sorted": 0,
                            "localPanelSettings": {
                                "xAxis": "_step",
                                "smoothingWeight": 0,
                                "smoothingType": "exponential",
                                "ignoreOutliers": False,
                                "xAxisActive": False,
                                "smoothingActive": False,
                            },
                        }
                    ],
                },
                "panelBankSectionConfig": {
                    "name": "Report Panels",
                    "isOpen": False,
                    "panels": [],
                    "type": "grid",
                    "flowConfig": {
                        "snapToColumns": True,
                        "columnsPerPage": 3,
                        "rowsPerPage": 2,
                        "gutterWidth": 16,
                        "boxWidth": 460,
                        "boxHeight": 300,
                    },
                    "sorted": 0,
                    "localPanelSettings": {
                        "xAxis": "_step",
                        "smoothingWeight": 0,
                        "smoothingType": "exponential",
                        "ignoreOutliers": False,
                        "xAxisActive": False,
                        "smoothingActive": False,
                    },
                },
                "customRunColors": {},
                "runSets": [],
                "openRunSet": 0,
                "name": "unused-name",
            },
        }

    @staticmethod
    def _default_runsets():
        return [RunSet()]

    @staticmethod
    def _default_panels():
        return []

    def _get_specific_keys_for_certain_plots(self, panels, setting=False):
        """
        Helper function to map names for certain plots
        """
        gen = self.runsets[0].pm_query_generator
        for p in panels:
            if isinstance(p, ParallelCoordinatesPlot):
                wandb.termlog(
                    "INFO: PCColumn metrics will be have special naming applied -- no change from you is required."
                )
                transform = gen.pc_front_to_back if setting else gen.pc_back_to_front
                if p.columns:
                    for col in p.columns:
                        col.metric = transform(col.metric)
        return panels


class Report(Base):
    def __init__(
        self,
        project,
        entity=None,
        title="Untitled Report",
        description="",
        width="readable",
        blocks=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._viewspec = self._default_viewspec()
        self._orig_viewspec = deepcopy(self._viewspec)

        self.project = project
        self.entity = coalesce(entity, wandb.Api().default_entity, "")
        self.title = title
        self.description = description
        self.width = width
        self.blocks = coalesce(blocks, [])

    project: str = Attr(json_path="viewspec.project.name")
    entity: str = Attr(json_path="viewspec.project.entityName")
    title: str = Attr(json_path="viewspec.displayName")
    description: str = Attr(json_path="viewspec.description")
    width: str = Attr(
        json_path="viewspec.spec.width",
        validators=[OneOf(["readable", "fixed", "fluid"])],
    )
    blocks: list = Attr(
        json_path="viewspec.spec.blocks",
        validators=[TypeValidator(Block, how="keys")],
    )

    @blocks.getter
    def blocks(self):
        json_path = self._get_path("blocks")
        block_specs = nested_get(self, json_path)
        blocks = []
        for bspec in block_specs:
            cls = block_mapping.get(bspec["type"], UnknownBlock)
            if cls is UnknownBlock:
                wandb.termwarn(
                    inspect.cleandoc(
                        f"""
                        UNKNOWN BLOCK DETECTED
                            This can happen if we have added new blocks, but you are using an older version of the SDK.
                            If your report is loading normally, you can safely ignore this message (but we recommend not touching UnknownPanel)
                            If you think this is an error, please file a bug report including your SDK version ({wandb.__version__}) and this spec ({bspec})
                        """
                    )
                )
            blocks.append(cls.from_json(bspec))
        return blocks[1:-1]  # accounts for hidden p blocks

    @blocks.setter
    def blocks(self, new_blocks):
        json_path = self._get_path("blocks")
        new_block_specs = (
            [P("").spec] + [b.spec for b in new_blocks] + [P("").spec]
        )  # hidden p blocks
        nested_set(self, json_path, new_block_specs)

    @staticmethod
    def _default_viewspec():
        return {
            "id": None,
            "name": None,
            "spec": {
                "version": 5,
                "panelSettings": {},
                "blocks": [],
                "width": "readable",
                "authors": [],
                "discussionThreads": [],
                "ref": {},
            },
        }

    @classmethod
    def from_json(cls, viewspec):
        obj = cls(project=viewspec["project"]["name"])
        obj._viewspec = viewspec
        obj._orig_viewspec = deepcopy(obj._viewspec)
        return obj

    @property
    def viewspec(self):
        return self._viewspec

    @property
    def modified(self) -> bool:
        return self._viewspec != self._orig_viewspec

    @property
    def spec(self) -> dict:
        return self._viewspec["spec"]

    @property
    def client(self) -> wandb.apis.public.RetryingClient:
        return wandb.Api().client

    @property
    def id(self) -> str:
        return self._viewspec["id"]

    @property
    def name(self) -> str:
        return self._viewspec["name"]

    @property
    def panel_grids(self) -> "LList[PanelGrid]":
        return [b for b in self.blocks if isinstance(b, PanelGrid)]

    @property
    def runsets(self) -> "LList[RunSet]":
        return [pg.runset for pg in self.panel_grids]

    @property
    def url(self) -> str:
        title = urllib.parse.quote(self.title.replace(" ", "-"))
        id = self.id.replace("=", "")
        return (
            f"{self.client.app_url}/{self.entity}/{self.project}/reports/{title}--{id}"
        )

    def save(self, draft: bool = False, clone: bool = False) -> "Report":
        if not self.modified:
            wandb.termwarn("Report has not been modified")

        # create project if not exists
        r = self.client.execute(
            CREATE_PROJECT, {"entityName": self.entity, "name": self.project}
        )

        r = self.client.execute(
            UPSERT_VIEW,
            variable_values={
                "id": None if clone or not self.id else self.id,
                "name": generate_name() if clone or not self.name else self.name,
                "entityName": self.entity,
                "projectName": self.project,
                "description": self.description,
                "displayName": self.title,
                "type": "runs/draft" if draft else "runs",
                "spec": json.dumps(self.spec),
            },
        )

        viewspec = r["upsertView"]["view"]
        viewspec["spec"] = json.loads(viewspec["spec"])
        if clone:
            return Report.from_json(viewspec)
        else:
            self._viewspec = viewspec
            return self

    def to_html(self, height: int = 1024, hidden: bool = False) -> str:
        """Generate HTML containing an iframe displaying this report"""
        try:
            url = self.url + "?jupyter=true"
            style = f"border:none;width:100%;height:{height}px;"
            prefix = ""
            if hidden:
                style += "display:none;"
                prefix = ipython.toggle_button("report")
            return prefix + f'<iframe src="{url}" style="{style}"></iframe>'
        except AttributeError:
            wandb.termlog("HTML repr will be available after you save the report!")

    def _repr_html_(self) -> str:
        return self.to_html()


class List(Base):
    @classmethod
    def from_json(cls, spec: dict) -> "Union[CheckedList, OrderedList, UnorderedList]":
        items = [
            item["children"][0]["children"][0]["text"] for item in spec["children"]
        ]
        checked = [item.get("checked") for item in spec["children"]]
        ordered = spec.get("ordered")

        # NAND: Either checked or ordered or neither (unordered), never both
        if all(x is None for x in checked):
            checked = None
        if checked is not None and ordered is not None:
            raise ValueError(
                "Lists can be checked, ordered or neither (unordered), but not both!"
            )

        if checked:
            return CheckedList(items, checked)
        elif ordered:
            return OrderedList(items)
        else:
            return UnorderedList(items)


class CheckedList(Block, List):
    def __init__(self, items, checked, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = items
        self.checked = checked

    items: list = Attr()
    checked: list = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "list",
            "children": [
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": [{"text": item}]}],
                    "checked": check,
                }
                for item, check in zip(self.items, self.checked)
            ],
        }


class OrderedList(Block, List):
    def __init__(self, items, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = items

    items: list = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "list",
            "ordered": True,
            "children": [
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": [{"text": item}]}],
                    "ordered": True,
                }
                for item in self.items
            ],
        }


class UnorderedList(Block, List):
    def __init__(self, items, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = items

    items: list = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "list",
            "children": [
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": [{"text": item}]}],
                }
                for item in self.items
            ],
        }


class Heading(Base):
    @classmethod
    def from_json(cls, spec: dict) -> "Union[H1,H2,H3]":
        level = spec["level"]
        text = spec["children"][0]["text"]

        level_mapping = {1: H1, 2: H2, 3: H3}

        if level not in level_mapping:
            raise ValueError(f"`level` must be one of {list(level_mapping.keys())}")

        return level_mapping[level](text)


class H1(Block, Heading):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: str = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "heading",
            "children": [{"text": self.text}],
            "level": 1,
        }


class H2(Block, Heading):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: str = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "heading",
            "children": [{"text": self.text}],
            "level": 2,
        }


class H3(Block, Heading):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: str = Attr()

    @property
    def spec(self) -> dict:
        return {
            "type": "heading",
            "children": [{"text": self.text}],
            "level": 3,
        }


class BlockQuote(Block):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "BlockQuote":
        text = spec["children"][0]["text"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {"type": "block-quote", "children": [{"text": self.text}]}


class CalloutBlock(Block):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: Union[str, list] = Attr()

    def __post_init__(self) -> None:
        if isinstance(self.text, str):
            self.text = self.text.split("\n")

    @classmethod
    def from_json(cls, spec: dict) -> "CalloutBlock":
        text = [child["children"][0]["text"] for child in spec["children"]]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "callout-block",
            "children": [
                {"type": "callout-line", "children": [{"text": text}]}
                for text in self.text
            ],
        }


class CodeBlock(Block):
    def __init__(self, code, language=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code = code
        self.language = coalesce(language, "python")

    code: Union[str, list] = Attr()
    language: str = Attr()

    def __post_init__(self) -> None:
        if isinstance(self.code, str):
            self.code = self.code.split("\n")

    @classmethod
    def from_json(cls, spec: dict) -> "CodeBlock":
        code = [child["children"][0]["text"] for child in spec["children"]]
        language = spec.get("language", "python")
        return cls(code, language)

    @property
    def spec(self) -> dict:
        language = self.language.lower()
        return {
            "type": "code-block",
            "children": [
                {
                    "type": "code-line",
                    "children": [{"text": text}],
                    "language": language,
                }
                for text in self.code
            ],
            "language": language,
        }


class MarkdownBlock(Block):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text
        if isinstance(self.text, list):
            self.text = "\n".join(self.text)

    text: Union[str, list] = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "MarkdownBlock":
        text = spec["content"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "markdown-block",
            "children": [{"text": ""}],
            "content": self.text,
        }


class LaTeXBlock(Block):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text
        if isinstance(self.text, list):
            self.text = "\n".join(self.text)

    text: Union[str, list] = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "LaTeXBlock":
        text = spec["content"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "latex",
            "children": [{"text": ""}],
            "content": self.text,
            "block": True,
        }


class Gallery(Block):
    def __init__(self, ids, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ids = ids

    ids: list = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "Gallery":
        ids = spec["ids"]
        return cls(ids)

    @classmethod
    def from_report_urls(cls, urls: LList[str]) -> "Gallery":
        ids = [url.split("--")[-1] for url in urls]
        return cls(ids)

    @property
    def spec(self) -> dict:
        return {"type": "gallery", "children": [{"text": ""}], "ids": self.ids}


class Image(Block):
    def __init__(self, url, caption, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.caption = caption

    url: str = Attr()
    caption: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "Image":
        url = spec["url"]
        caption = spec["children"][0]["text"] if spec.get("hasCaption") else None
        return cls(url, caption)

    @property
    def spec(self) -> dict:
        if self.caption:
            return {
                "type": "image",
                "children": [{"text": self.caption}],
                "url": self.url,
                "hasCaption": True,
            }
        else:
            return {"type": "image", "children": [{"text": ""}], "url": self.url}


class WeaveBlock(Block):
    def __init__(self, spec, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spec = spec

    spec: dict = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "WeaveBlock":
        return cls(spec)


class HorizontalRule(Block):
    @classmethod
    def from_json(cls, spec: dict) -> "HorizontalRule":
        return cls()

    @property
    def spec(self):
        return {"type": "horizontal-rule", "children": [{"text": ""}]}


class TableOfContents(Block):
    @classmethod
    def from_json(cls, spec: dict) -> "TableOfContents":
        return cls()

    @property
    def spec(self) -> dict:
        return {"type": "table-of-contents", "children": [{"text": ""}]}


class SoundCloud(Block):
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

    url: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "SoundCloud":
        quoted_url = spec["html"].split("url=")[-1].split("&show_artwork")[0]
        url = urllib.parse.unquote(quoted_url)
        return cls(url)

    @property
    def spec(self) -> dict:
        quoted_url = urllib.parse.quote(self.url)
        return {
            "type": "soundcloud",
            "html": f'<iframe width="100%" height="400" scrolling="no" frameborder="no" src="https://w.soundcloud.com/player/?visual=true&url={quoted_url}&show_artwork=true"></iframe>',
            "children": [{"text": ""}],
        }


class Twitter(Block):
    def __init__(self, embed_html, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.embed_html = embed_html
        if self.embed_html:
            pattern = r" <script[\s\S]+?/script>"
            self.embed_html = re.sub(pattern, "\n", self.embed_html)

    embed_html: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "Twitter":
        embed_html = spec["html"]
        return cls(embed_html)

    @property
    def spec(self) -> dict:
        return {"type": "twitter", "html": self.embed_html, "children": [{"text": ""}]}


class Spotify(Block):
    def __init__(self, spotify_id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spotify_id = spotify_id

    spotify_id: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "Spotify":
        return cls(spec["spotifyID"])

    @classmethod
    def from_url(cls, url: str) -> "Spotify":
        spotify_id = url.split("/")[-1].split("?")[0]
        return cls(spotify_id)

    @property
    def spec(self) -> dict:
        return {
            "type": "spotify",
            "spotifyType": "track",
            "spotifyID": self.spotify_id,
            "children": [{"text": ""}],
        }


class Video(Block):
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

    url: str = Attr()

    @classmethod
    def from_json(cls, spec: dict) -> "Video":
        return cls(spec["url"])

    @property
    def spec(self) -> dict:
        return {
            "type": "video",
            "url": self.url,
            "children": [{"text": ""}],
        }


class InlineLaTeX(Base):
    def __init__(self, latex, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.latex = latex

    latex: str = Attr()

    @property
    def spec(self) -> dict:
        return {"type": "latex", "children": [{"text": ""}], "content": self.latex}


class InlineCode(Base):
    def __init__(self, code, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code = code

    code: str = Attr()

    @property
    def spec(self) -> dict:
        return {"text": self.code, "inlineCode": True}


class P(Block):
    def __init__(self, text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    text: Union[str, InlineLaTeX, InlineCode, list] = Attr()

    @classmethod
    def from_json(cls, spec):
        if isinstance(spec["children"], str):
            text = spec["children"]
        else:
            text = []
            for elem in spec["children"]:
                if elem.get("type") == "latex":
                    text.append(InlineLaTeX(elem["content"]))
                elif elem.get("inlineCode"):
                    text.append(InlineCode(elem["text"]))
                else:
                    text.append(elem["text"])

        if not isinstance(text, list):
            text = [text]
        return cls(text)

    @property
    def spec(self) -> dict:
        if isinstance(self.text, list):
            content = [
                t.spec if not isinstance(t, str) else {"text": t} for t in self.text
            ]
        else:
            content = [{"text": self.text}]

        return {"type": "paragraph", "children": content}


block_mapping = {
    "block-quote": BlockQuote,
    "callout-block": CalloutBlock,
    "code-block": CodeBlock,
    "gallery": Gallery,
    "heading": Heading,
    "horizontal-rule": HorizontalRule,
    "image": Image,
    "latex": LaTeXBlock,
    "list": List,
    "markdown-block": MarkdownBlock,
    "panel-grid": PanelGrid,
    "paragraph": P,
    "table-of-contents": TableOfContents,
    "weave-panel": WeaveBlock,
    "video": Video,
    "spotify": Spotify,
    "twitter": Twitter,
    "soundcloud": SoundCloud,
}

panel_mapping = {
    # Panels with config
    "Run History Line Plot": LinePlot,
    "Scatter Plot": ScatterPlot,
    "Bar Chart": BarPlot,
    "Scalar Chart": ScalarChart,
    "Code Comparer": CodeComparer,
    "Parallel Coordinates Plot": ParallelCoordinatesPlot,
    "Parameter Importance": ParameterImportancePlot,
    "Run Comparer": RunComparer,
    "Media Browser": MediaBrowser,
    "Markdown Panel": MarkdownPanel,
    # Panels with no config
    "Confusion Matrix": ConfusionMatrix,
    "Data Frame Table": DataFrames,
    "Multi Run Table": MultiRunTable,
    "Vega": Vega,
    "Vega2": Vega2,
    "Vega3": Vega3,
    "Weave": WeavePanel,
}



import logging
from typing import Any, Callable, Dict


log = logging.getLogger(__name__)


def _import_sweep_scheduler() -> Any:
    from .scheduler_sweep import SweepScheduler

    return SweepScheduler


_WANDB_SCHEDULERS: Dict[str, Callable] = {
    "sweep": _import_sweep_scheduler,
}


def load_scheduler(scheduler_name: str) -> Any:

    scheduler_name = scheduler_name.lower()
    if scheduler_name not in _WANDB_SCHEDULERS:
        raise ValueError(
            f"The `scheduler_name` argument must be one of "
            f"{list(_WANDB_SCHEDULERS.keys())}, got: {scheduler_name}"
        )

    log.warn(f"Loading dependencies for Scheduler of type: {scheduler_name}")
    import_func = _WANDB_SCHEDULERS[scheduler_name]
    return import_func()


__all__ = [
    "load_scheduler",
]



from abc import ABC, abstractmethod
from typing import TypeVar, Union

LINEPLOT_STYLES = ["line", "stacked-area", "pct-area", None]
BARPLOT_STYLES = ["bar", "boxplot", "violin", None]
FONT_SIZES = ["small", "medium", "large", "auto", None]
LEGEND_POSITIONS = ["north", "south", "east", "west", None]
LEGEND_ORIENTATIONS = ["horizontal", "vertical", None]
AGGFUNCS = ["mean", "min", "max", "median", "sum", "samples", None]
RANGEFUNCS = ["minmax", "stddev", "stderr", "none", "samples", None]
MARKS = ["solid", "dashed", "dotted", "dotdash", "dotdotdash", None]
TIMESTEPS = ["seconds", "minutes", "hours", "days", None]
SMOOTHING_TYPES = ["exponential", "gaussian", "average", "none", None]
CODE_COMPARE_DIFF = ["split", "unified", None]


UNDEFINED_TYPE = TypeVar("UNDEFINED_TYPE")


class Validator(ABC):
    def __init__(self, how=None):
        self.how = how

    @abstractmethod
    def call(self, attr_name, value):
        pass

    def __call__(self, attr, value):
        attr_name = attr.name
        if value is None and self.how in {"keys", "values"}:
            return
        if self.how == "keys":
            attr_name += " keys"
            for v in value:
                self.call(attr_name, v)
        elif self.how == "values":
            attr_name += " values"
            for v in value.values():
                self.call(attr_name, v)
        else:
            attr_name += " object"
            self.call(attr_name, value)


class TypeValidator(Validator):
    def __init__(self, attr_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_type = attr_type
        try:
            origin = attr_type.__origin__
            subtypes = attr_type.__args__
        except AttributeError:  # normal types
            self.attr_type = (attr_type,)
        else:
            if origin is Union:
                self.attr_type = subtypes
            else:
                raise TypeError(f"{attr_type} is not currently supported.")

    def call(self, attr_name, value):
        if not isinstance(value, self.attr_type):
            raise TypeError(
                f"{attr_name} must be of type {self.attr_type!r} (got {type(value)!r})"
            )


class OneOf(Validator):
    def __init__(self, options, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = options

    def call(self, attr_name, value):
        if value not in self.options:
            raise ValueError(
                f"{attr_name} must be one of {self.options!r} (got {value!r})"
            )


class Length(Validator):
    def __init__(self, k, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.k = k

    def call(self, attr_name, value):
        if len(value) != self.k:
            raise ValueError(
                f"{attr_name} must have exactly {self.k} elements (got {len(value)!r}, elems: {value!r})"
            )


class Between(Validator):
    def __init__(self, lb, ub, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lb = lb
        self.ub = ub

    def call(self, attr_name, value):
        if not self.lb <= value <= self.ub:
            raise ValueError(
                f"{attr_name} must be between [{self.lb}, {self.ub}] inclusive (got {value})"
            )


class OrderString(TypeValidator):
    def __init__(self):
        super().__init__(attr_type=str)

    def call(self, attr_name, value):
        super().call(attr_name, value)

        if value[0] not in {"+", "-"}:
            raise ValueError(
                f'{attr_name} must be prefixed with "+" or "-" to indicate ascending or descending order'
            )


class LayoutDict(Validator):
    def call(self, attr_name, value):
        if set(value.keys()) != {"x", "y", "w", "h"}:
            raise ValueError(
                f"{attr_name} must be a dict containing exactly the keys `x`, y`, `w`, `h`"
            )
        for k, v in value.items():
            if not isinstance(v, int):
                raise ValueError(
                    f"{attr_name} key `{k}` must be of type {int} (got {type(v)!r})"
                )



import hashlib
import os
import platform
import shutil
from typing import cast, Optional, Sequence, Type, TYPE_CHECKING, Union

import wandb
from wandb import util
from wandb._globals import _datatypes_callback

from .wb_value import WBValue

if TYPE_CHECKING:  # pragma: no cover
    import numpy as np  # type: ignore
    from wandb.apis.public import Artifact as PublicArtifact

    from ...wandb_artifacts import Artifact as LocalArtifact
    from ...wandb_run import Run as LocalRun


SYS_PLATFORM = platform.system()


def _wb_filename(
    key: Union[str, int], step: Union[str, int], id: Union[str, int], extension: str
) -> str:
    return f"{str(key)}_{str(step)}_{str(id)}{extension}"


class Media(WBValue):
    """A WBValue that we store as a file outside JSON and show in a media panel
    on the front end.

    If necessary, we move or copy the file into the Run's media directory so
    that it gets uploaded.
    """

    _path: Optional[str]
    _run: Optional["LocalRun"]
    _caption: Optional[str]
    _is_tmp: Optional[bool]
    _extension: Optional[str]
    _sha256: Optional[str]
    _size: Optional[int]

    def __init__(self, caption: Optional[str] = None) -> None:
        super().__init__()
        self._path = None
        # The run under which this object is bound, if any.
        self._run = None
        self._caption = caption

    def _set_file(
        self, path: str, is_tmp: bool = False, extension: Optional[str] = None
    ) -> None:
        self._path = path
        self._is_tmp = is_tmp
        self._extension = extension
        assert extension is None or path.endswith(
            extension
        ), 'Media file extension "{}" must occur at the end of path "{}".'.format(
            extension, path
        )

        with open(self._path, "rb") as f:
            self._sha256 = hashlib.sha256(f.read()).hexdigest()
        self._size = os.path.getsize(self._path)

    @classmethod
    def get_media_subdir(cls: Type["Media"]) -> str:
        raise NotImplementedError

    @staticmethod
    def captions(
        media_items: Sequence["Media"],
    ) -> Union[bool, Sequence[Optional[str]]]:
        if media_items[0]._caption is not None:
            return [m._caption for m in media_items]
        else:
            return False

    def is_bound(self) -> bool:
        return self._run is not None

    def file_is_set(self) -> bool:
        return self._path is not None and self._sha256 is not None

    def bind_to_run(
        self,
        run: "LocalRun",
        key: Union[int, str],
        step: Union[int, str],
        id_: Optional[Union[int, str]] = None,
        ignore_copy_err: Optional[bool] = None,
    ) -> None:
        """Bind this object to a particular Run.

        Calling this function is necessary so that we have somewhere specific to
        put the file associated with this object, from which other Runs can
        refer to it.
        """
        assert self.file_is_set(), "bind_to_run called before _set_file"

        if SYS_PLATFORM == "Windows" and not util.check_windows_valid_filename(key):
            raise ValueError(
                f"Media {key} is invalid. Please remove invalid filename characters"
            )

        # The following two assertions are guaranteed to pass
        # by definition file_is_set, but are needed for
        # mypy to understand that these are strings below.
        assert isinstance(self._path, str)
        assert isinstance(self._sha256, str)

        assert run is not None, 'Argument "run" must not be None.'
        self._run = run

        if self._extension is None:
            _, extension = os.path.splitext(os.path.basename(self._path))
        else:
            extension = self._extension

        if id_ is None:
            id_ = self._sha256[:20]

        file_path = _wb_filename(key, step, id_, extension)
        media_path = os.path.join(self.get_media_subdir(), file_path)
        new_path = os.path.join(self._run.dir, media_path)
        util.mkdir_exists_ok(os.path.dirname(new_path))

        if self._is_tmp:
            shutil.move(self._path, new_path)
            self._path = new_path
            self._is_tmp = False
            _datatypes_callback(media_path)
        else:
            try:
                shutil.copy(self._path, new_path)
            except shutil.SameFileError as e:
                if not ignore_copy_err:
                    raise e
            self._path = new_path
            _datatypes_callback(media_path)

    def to_json(self, run: Union["LocalRun", "LocalArtifact"]) -> dict:
        """Serializes the object into a JSON blob, using a run or artifact to store additional data. If `run_or_artifact`
        is a wandb.Run then `self.bind_to_run()` must have been previously been called.

        Args:
            run_or_artifact (wandb.Run | wandb.Artifact): the Run or Artifact for which this object should be generating
            JSON for - this is useful to to store additional data if needed.

        Returns:
            dict: JSON representation
        """
        # NOTE: uses of Audio in this class are a temporary hack -- when Ref support moves up
        # into Media itself we should get rid of them
        from wandb.data_types import Audio

        json_obj = {}
        if isinstance(run, wandb.wandb_sdk.wandb_run.Run):
            json_obj.update(
                {
                    "_type": "file",  # TODO(adrian): This isn't (yet) a real media type we support on the frontend.
                    "sha256": self._sha256,
                    "size": self._size,
                }
            )
            artifact_entry_url = self._get_artifact_entry_ref_url()
            if artifact_entry_url is not None:
                json_obj["artifact_path"] = artifact_entry_url
            artifact_entry_latest_url = self._get_artifact_entry_latest_ref_url()
            if artifact_entry_latest_url is not None:
                json_obj["_latest_artifact_path"] = artifact_entry_latest_url

            if artifact_entry_url is None or self.is_bound():
                assert (
                    self.is_bound()
                ), "Value of type {} must be bound to a run with bind_to_run() before being serialized to JSON.".format(
                    type(self).__name__
                )

                assert (
                    self._run is run
                ), "We don't support referring to media files across runs."

                # The following two assertions are guaranteed to pass
                # by definition is_bound, but are needed for
                # mypy to understand that these are strings below.
                assert isinstance(self._path, str)
                json_obj["path"] = util.to_forward_slash_path(
                    os.path.relpath(self._path, self._run.dir)
                )

        elif isinstance(run, wandb.wandb_sdk.wandb_artifacts.Artifact):
            if self.file_is_set():
                # The following two assertions are guaranteed to pass
                # by definition of the call above, but are needed for
                # mypy to understand that these are strings below.
                assert isinstance(self._path, str)
                assert isinstance(self._sha256, str)
                artifact = run  # Checks if the concrete image has already been added to this artifact
                name = artifact.get_added_local_path_name(self._path)
                if name is None:
                    if self._is_tmp:
                        name = os.path.join(
                            self.get_media_subdir(), os.path.basename(self._path)
                        )
                    else:
                        # If the files is not temporary, include the first 8 characters of the file's SHA256 to
                        # avoid name collisions. This way, if there are two images `dir1/img.png` and `dir2/img.png`
                        # we end up with a unique path for each.
                        name = os.path.join(
                            self.get_media_subdir(),
                            self._sha256[:20],
                            os.path.basename(self._path),
                        )

                    # if not, check to see if there is a source artifact for this object
                    if (
                        self._artifact_source
                        is not None
                        # and self._artifact_source.artifact != artifact
                    ):
                        default_root = self._artifact_source.artifact._default_root()
                        # if there is, get the name of the entry (this might make sense to move to a helper off artifact)
                        if self._path.startswith(default_root):
                            name = self._path[len(default_root) :]
                            name = name.lstrip(os.sep)

                        # Add this image as a reference
                        path = self._artifact_source.artifact.get_path(name)
                        artifact.add_reference(path.ref_url(), name=name)
                    elif isinstance(self, Audio) and Audio.path_is_reference(
                        self._path
                    ):
                        artifact.add_reference(self._path, name=name)
                    else:
                        entry = artifact.add_file(
                            self._path, name=name, is_tmp=self._is_tmp
                        )
                        name = entry.path

                json_obj["path"] = name
                json_obj["sha256"] = self._sha256
            json_obj["_type"] = self._log_type
        return json_obj

    @classmethod
    def from_json(
        cls: Type["Media"], json_obj: dict, source_artifact: "PublicArtifact"
    ) -> "Media":
        """Likely will need to override for any more complicated media objects"""
        return cls(source_artifact.get_path(json_obj["path"]).download())

    def __eq__(self, other: object) -> bool:
        """Likely will need to override for any more complicated media objects"""
        return (
            isinstance(other, self.__class__)
            and hasattr(self, "_sha256")
            and hasattr(other, "_sha256")
            and self._sha256 == other._sha256
        )


class BatchableMedia(Media):
    """Parent class for Media we treat specially in batches, like images and
    thumbnails.

    Apart from images, we just use these batches to help organize files by name
    in the media directory.
    """

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def seq_to_json(
        cls: Type["BatchableMedia"],
        seq: Sequence["BatchableMedia"],
        run: "LocalRun",
        key: str,
        step: Union[int, str],
    ) -> dict:
        raise NotImplementedError


def _numpy_arrays_to_lists(
    payload: Union[dict, Sequence, "np.ndarray"]
) -> Union[Sequence, dict, str, int, float, bool]:
    # Casts all numpy arrays to lists so we don't convert them to histograms, primarily for Plotly

    if isinstance(payload, dict):
        res = {}
        for key, val in payload.items():
            res[key] = _numpy_arrays_to_lists(val)
        return res
    elif isinstance(payload, Sequence) and not isinstance(payload, str):
        return [_numpy_arrays_to_lists(v) for v in payload]
    elif util.is_numpy_array(payload):
        if TYPE_CHECKING:
            payload = cast("np.ndarray", payload)
        return [
            _numpy_arrays_to_lists(v)
            for v in (payload.tolist() if payload.ndim > 0 else [payload.tolist()])
        ]
    # Protects against logging non serializable objects
    elif isinstance(payload, Media):
        return str(payload.__class__.__name__)
    return payload



import numbers
from typing import Optional, Type, TYPE_CHECKING, Union

import wandb
from wandb import util
from wandb.util import has_num

from ..base_types.json_metadata import JSONMetadata

if TYPE_CHECKING:  # pragma: no cover
    from wandb.apis.public import Artifact as PublicArtifact

    from ...wandb_artifacts import Artifact as LocalArtifact
    from ...wandb_run import Run as LocalRun


class BoundingBoxes2D(JSONMetadata):
    """Format images with 2D bounding box overlays for logging to W&B.

    Arguments:
        val: (dictionary) A dictionary of the following form:
            box_data: (list of dictionaries) One dictionary for each bounding box, containing:
                position: (dictionary) the position and size of the bounding box, in one of two formats
                    Note that boxes need not all use the same format.
                    {"minX", "minY", "maxX", "maxY"}: (dictionary) A set of coordinates defining
                        the upper and lower bounds of the box (the bottom left and top right corners)
                    {"middle", "width", "height"}: (dictionary) A set of coordinates defining the
                        center and dimensions of the box, with "middle" as a list [x, y] for the
                        center point and "width" and "height" as numbers
                domain: (string) One of two options for the bounding box coordinate domain
                    null: By default, or if no argument is passed, the coordinate domain
                        is assumed to be relative to the original image, expressing this box as a fraction
                        or percentage of the original image. This means all coordinates and dimensions
                        passed into the "position" argument are floating point numbers between 0 and 1.
                    "pixel": (string literal) The coordinate domain is set to the pixel space. This means all
                        coordinates and dimensions passed into "position" are integers within the bounds
                        of the image dimensions.
                class_id: (integer) The class label id for this box
                scores: (dictionary of string to number, optional) A mapping of named fields
                        to numerical values (float or int), can be used for filtering boxes in the UI
                        based on a range of values for the corresponding field
                box_caption: (string, optional) A string to be displayed as the label text above this
                        box in the UI, often composed of the class label, class name, and/or scores

            class_labels: (dictionary, optional) A map of integer class labels to their readable class names

        key: (string)
            The readable name or id for this set of bounding boxes (e.g. predictions, ground_truth)

    Examples:
        ### Log bounding boxes for a single image
        <!--yeadoc-test:boundingbox-2d-->
        ```python
        import numpy as np
        import wandb

        wandb.init()
        image = np.random.randint(low=0, high=256, size=(200, 300, 3))

        class_labels = {
            0: "person",
            1: "car",
            2: "road",
            3: "building"
        }

        img = wandb.Image(image, boxes={
            "predictions": {
                "box_data": [
                    {
                        # one box expressed in the default relative/fractional domain
                        "position": {
                            "minX": 0.1,
                            "maxX": 0.2,
                            "minY": 0.3,
                            "maxY": 0.4
                        },
                        "class_id" : 1,
                        "box_caption": class_labels[1],
                        "scores" : {
                            "acc": 0.2,
                            "loss": 1.2
                        }
                    },
                    {
                        # another box expressed in the pixel domain
                        "position": {
                            "middle": [150, 20],
                            "width": 68,
                            "height": 112
                        },
                        "domain" : "pixel",
                        "class_id" : 3,
                        "box_caption": "a building",
                        "scores" : {
                            "acc": 0.5,
                            "loss": 0.7
                        }
                    },
                    # Log as many boxes an as needed
                ],
                "class_labels": class_labels
            }
        })

        wandb.log({"driving_scene": img})
        ```

        ### Log a bounding box overlay to a Table
        <!--yeadoc-test:bb2d-image-with-labels-->
        ```python

        import numpy as np
        import wandb

        wandb.init()
        image = np.random.randint(low=0, high=256, size=(200, 300, 3))

        class_labels = {
            0: "person",
            1: "car",
            2: "road",
            3: "building"
        }

        class_set = wandb.Classes([
            {"name" : "person", "id" : 0},
            {"name" : "car", "id" : 1},
            {"name" : "road", "id" : 2},
            {"name" : "building", "id" : 3}
        ])

        img = wandb.Image(image, boxes={
            "predictions": {
                "box_data": [
                    {
                        # one box expressed in the default relative/fractional domain
                        "position": {
                            "minX": 0.1,
                            "maxX": 0.2,
                            "minY": 0.3,
                            "maxY": 0.4
                        },
                        "class_id" : 1,
                        "box_caption": class_labels[1],
                        "scores" : {
                            "acc": 0.2,
                            "loss": 1.2
                        }
                    },
                    {
                        # another box expressed in the pixel domain
                        "position": {
                            "middle": [150, 20],
                            "width": 68,
                            "height": 112
                        },
                        "domain" : "pixel",
                        "class_id" : 3,
                        "box_caption": "a building",
                        "scores" : {
                            "acc": 0.5,
                            "loss": 0.7
                        }
                    },
                    # Log as many boxes an as needed
                ],
                "class_labels": class_labels
            }
        }, classes=class_set)

        table = wandb.Table(columns=["image"])
        table.add_data(img)
        wandb.log({"driving_scene": table})
        ```
    """

    _log_type = "bounding-boxes"
    # TODO: when the change is made to have this produce a dict with a _type, define
    # it here as _log_type, associate it in to_json

    def __init__(self, val: dict, key: str) -> None:
        """
        Arguments:
            val: (dictionary) A dictionary of the following form:
                box_data: (list of dictionaries) One dictionary for each bounding box, containing:
                    position: (dictionary) the position and size of the bounding box, in one of two formats
                        Note that boxes need not all use the same format.
                        {"minX", "minY", "maxX", "maxY"}: (dictionary) A set of coordinates defining
                            the upper and lower bounds of the box (the bottom left and top right corners)
                        {"middle", "width", "height"}: (dictionary) A set of coordinates defining the
                            center and dimensions of the box, with "middle" as a list [x, y] for the
                            center point and "width" and "height" as numbers
                    domain: (string) One of two options for the bounding box coordinate domain
                        null: By default, or if no argument is passed, the coordinate domain
                            is assumed to be relative to the original image, expressing this box as a fraction
                            or percentage of the original image. This means all coordinates and dimensions
                            passed into the "position" argument are floating point numbers between 0 and 1.
                        "pixel": (string literal) The coordinate domain is set to the pixel space. This means all
                            coordinates and dimensions passed into "position" are integers within the bounds
                            of the image dimensions.
                    class_id: (integer) The class label id for this box
                    scores: (dictionary of string to number, optional) A mapping of named fields
                            to numerical values (float or int), can be used for filtering boxes in the UI
                            based on a range of values for the corresponding field
                    box_caption: (string, optional) A string to be displayed as the label text above this
                            box in the UI, often composed of the class label, class name, and/or scores

                class_labels: (dictionary, optional) A map of integer class labels to their readable class names

            key: (string)
                The readable name or id for this set of bounding boxes (e.g. predictions, ground_truth)
        """
        super().__init__(val)
        self._val = val["box_data"]
        self._key = key
        # Add default class mapping
        if "class_labels" not in val:
            np = util.get_module(
                "numpy", required="Bounding box support requires numpy"
            )
            classes = (
                np.unique(list(box["class_id"] for box in val["box_data"]))
                .astype(np.int32)
                .tolist()
            )
            class_labels = {c: "class_" + str(c) for c in classes}
            self._class_labels = class_labels
        else:
            self._class_labels = val["class_labels"]

    def bind_to_run(
        self,
        run: "LocalRun",
        key: Union[int, str],
        step: Union[int, str],
        id_: Optional[Union[int, str]] = None,
        ignore_copy_err: Optional[bool] = None,
    ) -> None:
        # bind_to_run key argument is the Image parent key
        # the self._key value is the mask's sub key
        super().bind_to_run(run, key, step, id_=id_, ignore_copy_err=ignore_copy_err)
        run._add_singleton(
            "bounding_box/class_labels",
            str(key) + "_wandb_delimeter_" + self._key,
            self._class_labels,
        )

    @classmethod
    def type_name(cls) -> str:
        return "boxes2D"

    def validate(self, val: dict) -> bool:
        # Optional argument
        if "class_labels" in val:
            for k, v in list(val["class_labels"].items()):
                if (not isinstance(k, numbers.Number)) or (not isinstance(v, str)):
                    raise TypeError(
                        "Class labels must be a dictionary of numbers to string"
                    )

        boxes = val["box_data"]
        if not isinstance(boxes, list):
            raise TypeError("Boxes must be a list")

        for box in boxes:
            # Required arguments
            error_str = "Each box must contain a position with: middle, width, and height or \
                    \nminX, maxX, minY, maxY."
            if "position" not in box:
                raise TypeError(error_str)
            else:
                valid = False
                if (
                    "middle" in box["position"]
                    and len(box["position"]["middle"]) == 2
                    and has_num(box["position"], "width")
                    and has_num(box["position"], "height")
                ):
                    valid = True
                elif (
                    has_num(box["position"], "minX")
                    and has_num(box["position"], "maxX")
                    and has_num(box["position"], "minY")
                    and has_num(box["position"], "maxY")
                ):
                    valid = True

                if not valid:
                    raise TypeError(error_str)

            # Optional arguments
            if ("scores" in box) and not isinstance(box["scores"], dict):
                raise TypeError("Box scores must be a dictionary")
            elif "scores" in box:
                for k, v in list(box["scores"].items()):
                    if not isinstance(k, str):
                        raise TypeError("A score key must be a string")
                    if not isinstance(v, numbers.Number):
                        raise TypeError("A score value must be a number")

            if ("class_id" in box) and not isinstance(box["class_id"], int):
                raise TypeError("A box's class_id must be an integer")

            # Optional
            if ("box_caption" in box) and not isinstance(box["box_caption"], str):
                raise TypeError("A box's caption must be a string")
        return True

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:

        if isinstance(run_or_artifact, wandb.wandb_sdk.wandb_run.Run):
            return super().to_json(run_or_artifact)
        elif isinstance(run_or_artifact, wandb.wandb_sdk.wandb_artifacts.Artifact):
            # TODO (tim): I would like to log out a proper dictionary representing this object, but don't
            # want to mess with the visualizations that are currently available in the UI. This really should output
            # an object with a _type key. Will need to push this change to the UI first to ensure backwards compat
            return self._val
        else:
            raise ValueError("to_json accepts wandb_run.Run or wandb_artifact.Artifact")

    @classmethod
    def from_json(
        cls: Type["BoundingBoxes2D"], json_obj: dict, source_artifact: "PublicArtifact"
    ) -> "BoundingBoxes2D":
        return cls({"box_data": json_obj}, "")



"""
static settings.
"""
from typing import Any, Dict, Iterable, Optional, Tuple, Union

SettingsDict = Dict[str, Union[str, float, Tuple, None]]


class SettingsStatic:
    # TODO(jhr): figure out how to share type defs with sdk/wandb_settings.py
    _offline: Optional[bool]
    _disable_stats: Optional[bool]
    _disable_meta: Optional[bool]
    _start_time: float
    _start_datetime: str
    _stats_pid: int
    _stats_sample_rate_seconds: float
    _stats_samples_to_average: int
    files_dir: str
    log_internal: str
    _internal_check_process: bool
    is_local: Optional[bool]
    _colab: Optional[bool]
    _jupyter: Optional[bool]
    _require_service: Optional[str]
    _live_policy_rate_limit: Optional[int]
    resume: Optional[str]
    program: Optional[str]
    silent: Optional[bool]
    email: Optional[str]

    # TODO(jhr): clean this up, it is only in SettingsStatic and not in Settings
    _log_level: int

    def __init__(self, d: "SettingsDict") -> None:
        object.__setattr__(self, "__dict__", d)

    def __setattr__(self, name: str, value: object) -> None:
        raise AttributeError("Error: SettingsStatic is a readonly object")

    def __setitem__(self, key: str, val: object) -> None:
        raise AttributeError("Error: SettingsStatic is a readonly object")

    def keys(self) -> "Iterable[str]":
        return self.__dict__.keys()

    def items(self) -> "Iterable[Any]":
        return self.__dict__.items()

    def __getitem__(self, key: str) -> "Any":
        return self.__dict__[key]

    def __str__(self) -> str:
        return str(self.__dict__)



"""Backend Sender - Send to internal process

Manage backend sender.

"""

import logging
from typing import Any, Optional
from typing import TYPE_CHECKING

import grpc
from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2_grpc as pbgrpc
from wandb.proto import wandb_telemetry_pb2 as tpb

from .interface import InterfaceBase
from .message_future import MessageFuture
from .message_future_poll import MessageFuturePoll


if TYPE_CHECKING:
    from ..wandb_run import Run


logger = logging.getLogger("wandb")


class InterfaceGrpc(InterfaceBase):

    _stub: Optional[pbgrpc.InternalServiceStub]
    _stream_id: Optional[str]

    def __init__(self) -> None:
        super().__init__()
        self._stub = None
        self._process_check = None
        self._stream_id = None

    def _hack_set_run(self, run: "Run") -> None:
        super()._hack_set_run(run)
        assert run._run_id
        self._stream_id = run._run_id

    def _connect(self, stub: pbgrpc.InternalServiceStub) -> None:
        self._stub = stub

    def _assign(self, record: Any) -> None:
        assert self._stream_id
        record._info.stream_id = self._stream_id

    def _communicate_check_version(
        self, check_version: pb.CheckVersionRequest
    ) -> Optional[pb.CheckVersionResponse]:
        assert self._stub
        self._assign(check_version)
        run_result = self._stub.CheckVersion(check_version)
        return run_result

    def _communicate_attach(
        self, attach: pb.AttachRequest
    ) -> Optional[pb.AttachResponse]:
        assert self._stub
        self._assign(attach)
        resp = self._stub.Attach(attach)
        return resp

    def _communicate_run(
        self, run: pb.RunRecord, timeout: int = None
    ) -> Optional[pb.RunUpdateResult]:
        assert self._stub
        self._assign(run)
        run_result = self._stub.RunUpdate(run)
        return run_result

    def _publish_run(self, run: pb.RunRecord) -> None:
        assert self._stub
        self._assign(run)
        _ = self._stub.RunUpdate(run)

    def _publish_config(self, cfg: pb.ConfigRecord) -> None:
        assert self._stub
        self._assign(cfg)
        _ = self._stub.Config(cfg)

    def _publish_metric(self, metric: pb.MetricRecord) -> None:
        assert self._stub
        self._assign(metric)
        _ = self._stub.Metric(metric)

    def _publish_summary(self, summary: pb.SummaryRecord) -> None:
        assert self._stub
        self._assign(summary)
        _ = self._stub.Summary(summary)

    def _communicate_get_summary(
        self, get_summary: pb.GetSummaryRequest
    ) -> Optional[pb.GetSummaryResponse]:
        assert self._stub
        self._assign(get_summary)
        try:
            resp = self._stub.GetSummary(get_summary)
        except grpc.RpcError as e:
            logger.info(f"GET SUMMARY TIMEOUT: {e}")
            resp = pb.GetSummaryResponse()
        return resp

    def _publish_telemetry(self, telem: tpb.TelemetryRecord) -> None:
        assert self._stub
        self._assign(telem)
        _ = self._stub.Telemetry(telem)

    def _publish_partial_history(
        self, partial_history: pb.PartialHistoryRequest
    ) -> None:
        assert self._stub
        self._assign(partial_history)
        _ = self._stub.PartialLog(partial_history)

    def _publish_history(self, history: pb.HistoryRecord) -> None:
        assert self._stub
        self._assign(history)
        _ = self._stub.Log(history)

    def _publish_preempting(self, preempt_rec: pb.RunPreemptingRecord) -> None:
        assert self._stub
        self._assign(preempt_rec)
        _ = self._stub.RunPreempting(preempt_rec)

    def _publish_output(self, outdata: pb.OutputRecord) -> None:
        assert self._stub
        self._assign(outdata)
        _ = self._stub.Output(outdata)

    def _communicate_shutdown(self) -> None:
        assert self._stub
        shutdown = pb.ShutdownRequest()
        self._assign(shutdown)
        _ = self._stub.Shutdown(shutdown)

    def _communicate_run_start(
        self, run_start: pb.RunStartRequest
    ) -> Optional[pb.RunStartResponse]:
        assert self._stub
        self._assign(run_start)
        try:
            run_start_response = self._stub.RunStart(run_start)
        except grpc.RpcError as e:
            logger.info(f"RUNSTART TIMEOUT: {e}")
            run_start_response = pb.RunStartResponse()
        return run_start_response

    def _publish_files(self, files: pb.FilesRecord) -> None:
        assert self._stub
        self._assign(files)
        _ = self._stub.Files(files)

    def _publish_artifact(self, proto_artifact: pb.ArtifactRecord) -> None:
        assert self._stub
        self._assign(proto_artifact)
        _ = self._stub.Artifact(proto_artifact)

    def _publish_link_artifact(self, link_artifact: pb.LinkArtifactRecord) -> None:
        assert self._stub
        self._assign(link_artifact)
        _ = self._stub.LinkArtifact(link_artifact)

    def _communicate_artifact(
        self, log_artifact: pb.LogArtifactRequest
    ) -> MessageFuture:
        art_send = pb.ArtifactSendRequest()
        art_send.artifact.CopyFrom(log_artifact.artifact)

        assert self._stub
        self._assign(art_send)
        art_send_resp = self._stub.ArtifactSend(art_send)

        xid = art_send_resp.xid
        future = MessageFuturePoll(self._future_poll_artifact, xid)
        return future

    def _future_poll_artifact(self, xid: str) -> Optional[pb.Result]:
        art_poll = pb.ArtifactPollRequest(xid=xid)

        assert self._stub
        self._assign(art_poll)
        art_poll_resp = self._stub.ArtifactPoll(art_poll)

        if not art_poll_resp.ready:
            return None

        # emulate log_artifact response for old _communicate_artifact() protocol
        result = pb.Result()
        result.response.log_artifact_response.artifact_id = art_poll_resp.artifact_id
        result.response.log_artifact_response.error_message = (
            art_poll_resp.error_message
        )
        return result

    def _communicate_artifact_send(
        self, art_send: pb.ArtifactSendRequest
    ) -> Optional[pb.ArtifactSendResponse]:
        # NOTE: Not used in grpc interface now
        raise NotImplementedError

    def _communicate_artifact_poll(
        self, art_send: pb.ArtifactPollRequest
    ) -> Optional[pb.ArtifactPollResponse]:
        # NOTE: Not used in grpc interface now
        raise NotImplementedError

    def _publish_artifact_done(self, artifact_done: pb.ArtifactDoneRequest) -> None:
        # NOTE: Not used in grpc interface now
        raise NotImplementedError

    def _communicate_status(
        self, status: pb.StatusRequest
    ) -> Optional[pb.StatusResponse]:
        assert self._stub
        self._assign(status)
        status_response = self._stub.Status(status)
        return status_response

    def _communicate_network_status(
        self, status: pb.NetworkStatusRequest
    ) -> Optional[pb.NetworkStatusResponse]:
        assert self._stub
        self._assign(status)
        # TODO: implement
        return None

    def _communicate_stop_status(
        self, status: pb.StopStatusRequest
    ) -> Optional[pb.StopStatusResponse]:
        assert self._stub
        self._assign(status)
        # TODO: implement
        return None

    def _publish_alert(self, alert: pb.AlertRecord) -> None:
        assert self._stub
        self._assign(alert)
        _ = self._stub.Alert(alert)

    def _publish_tbdata(self, tbrecord: pb.TBRecord) -> None:
        assert self._stub
        self._assign(tbrecord)
        _ = self._stub.TBSend(tbrecord)

    def _publish_exit(self, exit_data: pb.RunExitRecord) -> None:
        assert self._stub
        self._assign(exit_data)
        _ = self._stub.RunExit(exit_data)
        return None

    def _communicate_poll_exit(
        self, poll_exit: pb.PollExitRequest
    ) -> Optional[pb.PollExitResponse]:
        assert self._stub
        self._assign(poll_exit)
        try:
            ret = self._stub.PollExit(poll_exit)
        except grpc.RpcError as e:
            logger.info(f"POLL EXIT TIMEOUT: {e}")
            ret = pb.PollExitResponse()
        return ret

    def _communicate_sampled_history(
        self, sampled_history: pb.SampledHistoryRequest
    ) -> Optional[pb.SampledHistoryResponse]:
        assert self._stub
        self._assign(sampled_history)
        ret = self._stub.SampledHistory(sampled_history)
        return ret

    def _publish_header(self, header: pb.HeaderRecord) -> None:
        assert self._stub
        # TODO: implement?

    def _publish_pause(self, pause: pb.PauseRequest) -> None:
        assert self._stub
        self._assign(pause)
        _ = self._stub.Pause(pause)

    def _publish_resume(self, resume: pb.ResumeRequest) -> None:
        assert self._stub
        self._assign(resume)
        _ = self._stub.Resume(resume)

    def join(self) -> None:
        super().join()



#!/usr/bin/env python

import os
import subprocess
from typing import Any, Dict, Optional

import grpc_tools  # type: ignore
from grpc_tools import protoc  # type: ignore


def generate_deprecated_class_definition() -> None:
    """
    Generate a class definition listing the deprecated features.
    This is to allow static checks to ensure that proper field names are used.
    """
    from wandb.proto.wandb_telemetry_pb2 import Deprecated  # type: ignore[import]
    deprecated_features = Deprecated.DESCRIPTOR.fields_by_name.keys()

    code: str = (
        "# Generated by wandb/proto/wandb_internal_codegen.py.  DO NOT EDIT!\n\n\n"
        "import sys\n\n\n"
        "if sys.version_info >= (3, 8):\n"
        "    from typing import Literal\n"
        "else:\n"
        "    from typing_extensions import Literal\n\n\n"
        "DEPRECATED_FEATURES = Literal[\n"
        + ",\n".join(f'    "{feature}"' for feature in deprecated_features) + "\n"
        + "]\n\n\n"
        "class Deprecated:\n"
        + "".join(
            [
                f'    {feature}: DEPRECATED_FEATURES = "{feature}"\n'
                for feature in deprecated_features
            ]
        )
    )
    with open("wandb/proto/wandb_deprecated.py", "w") as f:
        f.write(code)


def get_pip_package_version(package_name: str) -> str:
    out = subprocess.check_output(("pip", "show", package_name))
    info: Dict[str, Any] = dict(
        [line.split(": ", 2) for line in out.decode().rstrip("\n").split("\n")]  # type: ignore[misc]
    )
    return info["Version"]


def get_requirements_version(requirements_file_name: str, package_name: str) -> Optional[str]:
    with open(requirements_file_name) as f:
        lines = f.readlines()
        for line in lines:
            tokens = line.strip().split("==")
            if tokens[0] == package_name:
                assert len(tokens) == 2, f"Package {package_name} not pinned"
                return tokens[1]
    return None


package: str = "grpcio-tools"
package_version = get_pip_package_version(package)
requirements_file: str = "../../requirements_build.txt"
requirements_version = get_requirements_version(requirements_file, package)
assert package_version == requirements_version, (
    f"Package {package} found={package_version} required={requirements_version}"
)

proto_root = os.path.join(os.path.dirname(grpc_tools.__file__), "_proto")
os.chdir("../..")
ret = protoc.main(
    (
        '',
        '-I', proto_root,
        '-I', '.',
        '--python_out=.',
        '--mypy_out=.',
        'wandb/proto/wandb_base.proto',
    )
)
assert not ret

ret = protoc.main(
    (
        '',
        '-I', proto_root,
        '-I', '.',
        '--python_out=.',
        '--mypy_out=.',
        'wandb/proto/wandb_internal.proto',
    )
)
assert not ret

ret = protoc.main(
    (
        '',
        '-I', proto_root,
        '-I', '.',
        '--python_out=.',
        '--mypy_out=.',
        'wandb/proto/wandb_telemetry.proto',
    )
)
assert not ret

ret = protoc.main(
    (
        '',
        '-I', proto_root,
        '-I', '.',
        '--python_out=.',
        '--grpc_python_out=.',
        '--mypy_out=.',
        '--mypy_grpc_out=.',
        'wandb/proto/wandb_server.proto',
    )
)
assert not ret

generate_deprecated_class_definition()



from .arguments_of_correct_type import ArgumentsOfCorrectType
from .default_values_of_correct_type import DefaultValuesOfCorrectType
from .fields_on_correct_type import FieldsOnCorrectType
from .fragments_on_composite_types import FragmentsOnCompositeTypes
from .known_argument_names import KnownArgumentNames
from .known_directives import KnownDirectives
from .known_fragment_names import KnownFragmentNames
from .known_type_names import KnownTypeNames
from .lone_anonymous_operation import LoneAnonymousOperation
from .no_fragment_cycles import NoFragmentCycles
from .no_undefined_variables import NoUndefinedVariables
from .no_unused_fragments import NoUnusedFragments
from .no_unused_variables import NoUnusedVariables
from .overlapping_fields_can_be_merged import OverlappingFieldsCanBeMerged
from .possible_fragment_spreads import PossibleFragmentSpreads
from .provided_non_null_arguments import ProvidedNonNullArguments
from .scalar_leafs import ScalarLeafs
from .unique_argument_names import UniqueArgumentNames
from .unique_fragment_names import UniqueFragmentNames
from .unique_input_field_names import UniqueInputFieldNames
from .unique_operation_names import UniqueOperationNames
from .unique_variable_names import UniqueVariableNames
from .variables_are_input_types import VariablesAreInputTypes
from .variables_in_allowed_position import VariablesInAllowedPosition

specified_rules = [
    UniqueOperationNames,
    LoneAnonymousOperation,
    KnownTypeNames,
    FragmentsOnCompositeTypes,
    VariablesAreInputTypes,
    ScalarLeafs,
    FieldsOnCorrectType,
    UniqueFragmentNames,
    KnownFragmentNames,
    NoUnusedFragments,
    PossibleFragmentSpreads,
    NoFragmentCycles,
    NoUndefinedVariables,
    NoUnusedVariables,
    KnownDirectives,
    KnownArgumentNames,
    UniqueArgumentNames,
    ArgumentsOfCorrectType,
    ProvidedNonNullArguments,
    DefaultValuesOfCorrectType,
    VariablesInAllowedPosition,
    OverlappingFieldsCanBeMerged,
    UniqueInputFieldNames,
    UniqueVariableNames
]

__all__ = [
    'ArgumentsOfCorrectType',
    'DefaultValuesOfCorrectType',
    'FieldsOnCorrectType',
    'FragmentsOnCompositeTypes',
    'KnownArgumentNames',
    'KnownDirectives',
    'KnownFragmentNames',
    'KnownTypeNames',
    'LoneAnonymousOperation',
    'NoFragmentCycles',
    'UniqueVariableNames',
    'NoUndefinedVariables',
    'NoUnusedFragments',
    'NoUnusedVariables',
    'OverlappingFieldsCanBeMerged',
    'PossibleFragmentSpreads',
    'ProvidedNonNullArguments',
    'ScalarLeafs',
    'UniqueArgumentNames',
    'UniqueFragmentNames',
    'UniqueInputFieldNames',
    'UniqueOperationNames',
    'VariablesAreInputTypes',
    'VariablesInAllowedPosition',
    'specified_rules'
]



import wandb
from wandb import util
from wandb.plots.utils import test_missing, test_types


def pr_curve(
    y_true=None,
    y_probas=None,
    labels=None,
    classes_to_plot=None,
    interp_size=21,
    title=None,
):
    """
    Computes the tradeoff between precision and recall for different thresholds.
    A high area under the curve represents both high recall and high precision,
    where high precision relates to a low false positive rate, and high recall
    relates to a low false negative rate. High scores for both show that the
    classifier is returning accurate results (high precision), and
    returning a majority of all positive results (high recall).
    PR curve is useful when the classes are very imbalanced.

    Arguments:
        y_true (arr): true sparse labels
        y_probas (arr): Target scores, can either be probability estimates, confidence
                         values, or non-thresholded measure of decisions.
                         shape: (*y_true.shape, num_classes)
        labels (list): Named labels for target variable (y). Makes plots easier to
                        read by replacing target values with corresponding index.
                        For example labels = ['dog', 'cat', 'owl'] all 0s are
                        replaced by 'dog', 1s by 'cat'.
        classes_to_plot (list): unique values of y_true to include in the plot
        interp_size (int): the recall values will be fixed to `interp_size` points
                            uniform on [0, 1] and the precision will be interpolated for
                            these recall values.

    Returns:
        Nothing. To see plots, go to your W&B run page then expand the 'media' tab
        under 'auto visualizations'.

    Example:
        ```
        wandb.log({"pr-curve": wandb.plot.pr_curve(y_true, y_probas, labels)})
        ```
    """
    np = util.get_module(
        "numpy",
        required="roc requires the numpy library, install with `pip install numpy`",
    )
    pd = util.get_module(
        "pandas",
        required="roc requires the pandas library, install with `pip install pandas`",
    )
    sklearn_metrics = util.get_module(
        "sklearn.metrics",
        "roc requires the scikit library, install with `pip install scikit-learn`",
    )
    sklearn_utils = util.get_module(
        "sklearn.utils",
        "roc requires the scikit library, install with `pip install scikit-learn`",
    )

    def _step(x):
        y = np.array(x)
        for i in range(1, len(y)):
            y[i] = max(y[i], y[i - 1])
        return y

    y_true = np.array(y_true)
    y_probas = np.array(y_probas)

    if not test_missing(y_true=y_true, y_probas=y_probas):
        return
    if not test_types(y_true=y_true, y_probas=y_probas):
        return

    classes = np.unique(y_true)
    if classes_to_plot is None:
        classes_to_plot = classes

    precision = dict()
    interp_recall = np.linspace(0, 1, interp_size)[::-1]
    indices_to_plot = np.where(np.isin(classes, classes_to_plot))[0]
    for i in indices_to_plot:
        if labels is not None and (
            isinstance(classes[i], int) or isinstance(classes[0], np.integer)
        ):
            class_label = labels[classes[i]]
        else:
            class_label = classes[i]

        cur_precision, cur_recall, _ = sklearn_metrics.precision_recall_curve(
            y_true, y_probas[:, i], pos_label=classes[i]
        )
        # smooth the precision (monotonically increasing)
        cur_precision = _step(cur_precision)

        # reverse order so that recall in ascending
        cur_precision = cur_precision[::-1]
        cur_recall = cur_recall[::-1]
        indices = np.searchsorted(cur_recall, interp_recall, side="left")
        precision[class_label] = cur_precision[indices]

    df = pd.DataFrame(
        {
            "class": np.hstack([[k] * len(v) for k, v in precision.items()]),
            "precision": np.hstack(list(precision.values())),
            "recall": np.tile(interp_recall, len(precision)),
        }
    )
    df = df.round(3)

    if len(df) > wandb.Table.MAX_ROWS:
        wandb.termwarn(
            "wandb uses only %d data points to create the plots." % wandb.Table.MAX_ROWS
        )
        # different sampling could be applied, possibly to ensure endpoints are kept
        df = sklearn_utils.resample(
            df,
            replace=False,
            n_samples=wandb.Table.MAX_ROWS,
            random_state=42,
            stratify=df["class"],
        ).sort_values(["precision", "recall", "class"])

    table = wandb.Table(dataframe=df)
    title = title or "Precision v. Recall"
    return wandb.plot_table(
        "wandb/area-under-curve/v0",
        table,
        {"x": "recall", "y": "precision", "class": "class"},
        {"title": title},
    )



from collections import defaultdict

from ..error import GraphQLError
from ..language import ast
from ..pyutils.ordereddict import OrderedDict
from ..type.definition import (GraphQLArgument, GraphQLEnumType,
                               GraphQLEnumValue, GraphQLField,
                               GraphQLInputObjectField, GraphQLInputObjectType,
                               GraphQLInterfaceType, GraphQLList,
                               GraphQLNonNull, GraphQLObjectType,
                               GraphQLScalarType, GraphQLUnionType)
from ..type.introspection import (__Directive, __DirectiveLocation,
                                  __EnumValue, __Field, __InputValue, __Schema,
                                  __Type, __TypeKind)
from ..type.scalars import (GraphQLBoolean, GraphQLFloat, GraphQLID,
                            GraphQLInt, GraphQLString)
from ..type.schema import GraphQLSchema
from .value_from_ast import value_from_ast


def extend_schema(schema, documentAST=None):
    """Produces a new schema given an existing schema and a document which may
    contain GraphQL type extensions and definitions. The original schema will
    remain unaltered.

    Because a schema represents a graph of references, a schema cannot be
    extended without effectively making an entire copy. We do not know until it's
    too late if subgraphs remain unchanged.

    This algorithm copies the provided schema, applying extensions while
    producing the copy. The original schema remains unaltered."""

    assert isinstance(
        schema, GraphQLSchema), 'Must provide valid GraphQLSchema'
    assert documentAST and isinstance(
        documentAST, ast.Document), 'Must provide valid Document AST'

    # Collect the type definitions and extensions found in the document.
    type_definition_map = {}
    type_extensions_map = defaultdict(list)

    for _def in documentAST.definitions:
        if isinstance(_def, (
            ast.ObjectTypeDefinition,
            ast.InterfaceTypeDefinition,
            ast.EnumTypeDefinition,
            ast.UnionTypeDefinition,
            ast.ScalarTypeDefinition,
            ast.InputObjectTypeDefinition,
        )):
            # Sanity check that none of the defined types conflict with the
            # schema's existing types.
            type_name = _def.name.value
            if schema.get_type(type_name):
                raise GraphQLError(
                    ('Type "{}" already exists in the schema. It cannot also ' +
                     'be defined in this type definition.').format(type_name),
                    [_def]
                )

            type_definition_map[type_name] = _def
        elif isinstance(_def, ast.TypeExtensionDefinition):
            # Sanity check that this type extension exists within the
            # schema's existing types.
            extended_type_name = _def.definition.name.value
            existing_type = schema.get_type(extended_type_name)
            if not existing_type:
                raise GraphQLError(
                    ('Cannot extend type "{}" because it does not ' +
                     'exist in the existing schema.').format(extended_type_name),
                    [_def.definition]
                )
            if not isinstance(existing_type, GraphQLObjectType):
                raise GraphQLError(
                    'Cannot extend non-object type "{}".'.format(
                        extended_type_name),
                    [_def.definition]
                )

            type_extensions_map[extended_type_name].append(_def)

    # Below are functions used for producing this schema that have closed over
    # this scope and have access to the schema, cache, and newly defined types.

    def get_type_from_def(type_def):
        type = _get_named_type(type_def.name)
        assert type, 'Invalid schema'
        return type

    def get_type_from_AST(astNode):
        type = _get_named_type(astNode.name.value)
        if not type:
            raise GraphQLError(
                ('Unknown type: "{}". Ensure that this type exists ' +
                 'either in the original schema, or is added in a type definition.').format(
                    astNode.name.value),
                [astNode]
            )
        return type

    # Given a name, returns a type from either the existing schema or an
    # added type.
    def _get_named_type(typeName):
        cached_type_def = type_def_cache.get(typeName)
        if cached_type_def:
            return cached_type_def

        existing_type = schema.get_type(typeName)
        if existing_type:
            type_def = extend_type(existing_type)
            type_def_cache[typeName] = type_def
            return type_def

        type_ast = type_definition_map.get(typeName)
        if type_ast:
            type_def = build_type(type_ast)
            type_def_cache[typeName] = type_def
            return type_def

    # Given a type's introspection result, construct the correct
    # GraphQLType instance.
    def extend_type(type):
        if isinstance(type, GraphQLObjectType):
            return extend_object_type(type)
        if isinstance(type, GraphQLInterfaceType):
            return extend_interface_type(type)
        if isinstance(type, GraphQLUnionType):
            return extend_union_type(type)
        return type

    def extend_object_type(type):
        return GraphQLObjectType(
            name=type.name,
            description=type.description,
            interfaces=lambda: extend_implemented_interfaces(type),
            fields=lambda: extend_field_map(type),
        )

    def extend_interface_type(type):
        return GraphQLInterfaceType(
            name=type.name,
            description=type.description,
            fields=lambda: extend_field_map(type),
            resolve_type=cannot_execute_client_schema,
        )

    def extend_union_type(type):
        return GraphQLUnionType(
            name=type.name,
            description=type.description,
            types=list(map(get_type_from_def, type.types)),
            resolve_type=cannot_execute_client_schema,
        )

    def extend_implemented_interfaces(type):
        interfaces = list(map(get_type_from_def, type.interfaces))

        # If there are any extensions to the interfaces, apply those here.
        extensions = type_extensions_map[type.name]
        for extension in extensions:
            for namedType in extension.definition.interfaces:
                interface_name = namedType.name.value
                if any([_def.name == interface_name for _def in interfaces]):
                    raise GraphQLError(
                        ('Type "{}" already implements "{}". ' +
                         'It cannot also be implemented in this type extension.').format(
                            type.name, interface_name),
                        [namedType]
                    )
                interfaces.append(get_type_from_AST(namedType))

        return interfaces

    def extend_field_map(type):
        new_field_map = OrderedDict()
        old_field_map = type.fields
        for field_name, field in old_field_map.items():
            new_field_map[field_name] = GraphQLField(
                extend_field_type(field.type),
                description=field.description,
                deprecation_reason=field.deprecation_reason,
                args=field.args,
                resolver=cannot_execute_client_schema,
            )

        # If there are any extensions to the fields, apply those here.
        extensions = type_extensions_map[type.name]
        for extension in extensions:
            for field in extension.definition.fields:
                field_name = field.name.value
                if field_name in old_field_map:
                    raise GraphQLError(
                        ('Field "{}.{}" already exists in the ' +
                         'schema. It cannot also be defined in this type extension.').format(
                            type.name, field_name),
                        [field]
                    )
                new_field_map[field_name] = GraphQLField(
                    build_field_type(field.type),
                    args=build_input_values(field.arguments),
                    resolver=cannot_execute_client_schema,
                )

        return new_field_map

    def extend_field_type(type):
        if isinstance(type, GraphQLList):
            return GraphQLList(extend_field_type(type.of_type))
        if isinstance(type, GraphQLNonNull):
            return GraphQLNonNull(extend_field_type(type.of_type))
        return get_type_from_def(type)

    def build_type(type_ast):
        _type_build = {
            ast.ObjectTypeDefinition: build_object_type,
            ast.InterfaceTypeDefinition: build_interface_type,
            ast.UnionTypeDefinition: build_union_type,
            ast.ScalarTypeDefinition: build_scalar_type,
            ast.EnumTypeDefinition: build_enum_type,
            ast.InputObjectTypeDefinition: build_input_object_type
        }
        func = _type_build.get(type(type_ast))
        if func:
            return func(type_ast)

    def build_object_type(type_ast):
        return GraphQLObjectType(
            type_ast.name.value,
            interfaces=lambda: build_implemented_interfaces(type_ast),
            fields=lambda: build_field_map(type_ast),
        )

    def build_interface_type(type_ast):
        return GraphQLInterfaceType(
            type_ast.name.value,
            fields=lambda: build_field_map(type_ast),
            resolve_type=cannot_execute_client_schema,
        )

    def build_union_type(type_ast):
        return GraphQLUnionType(
            type_ast.name.value,
            types=list(map(get_type_from_AST, type_ast.types)),
            resolve_type=cannot_execute_client_schema,
        )

    def build_scalar_type(type_ast):
        return GraphQLScalarType(
            type_ast.name.value,
            serialize=lambda *args, **kwargs: None,
            # Note: validation calls the parse functions to determine if a
            # literal value is correct. Returning null would cause use of custom
            # scalars to always fail validation. Returning false causes them to
            # always pass validation.
            parse_value=lambda *args, **kwargs: False,
            parse_literal=lambda *args, **kwargs: False,
        )

    def build_enum_type(type_ast):
        return GraphQLEnumType(
            type_ast.name.value,
            values={v.name.value: GraphQLEnumValue() for v in type_ast.values},
        )

    def build_input_object_type(type_ast):
        return GraphQLInputObjectType(
            type_ast.name.value,
            fields=lambda: build_input_values(
                type_ast.fields, GraphQLInputObjectField),
        )

    def build_implemented_interfaces(type_ast):
        return list(map(get_type_from_AST, type_ast.interfaces))

    def build_field_map(type_ast):
        return {
            field.name.value: GraphQLField(
                build_field_type(field.type),
                args=build_input_values(field.arguments),
                resolver=cannot_execute_client_schema,
            ) for field in type_ast.fields
        }

    def build_input_values(values, input_type=GraphQLArgument):
        input_values = OrderedDict()
        for value in values:
            type = build_field_type(value.type)
            input_values[value.name.value] = input_type(
                type,
                default_value=value_from_ast(value.default_value, type)
            )
        return input_values

    def build_field_type(type_ast):
        if isinstance(type_ast, ast.ListType):
            return GraphQLList(build_field_type(type_ast.type))
        if isinstance(type_ast, ast.NonNullType):
            return GraphQLNonNull(build_field_type(type_ast.type))
        return get_type_from_AST(type_ast)

    # If this document contains no new types, then return the same unmodified
    # GraphQLSchema instance.
    if not type_extensions_map and not type_definition_map:
        return schema

    # A cache to use to store the actual GraphQLType definition objects by name.
    # Initialize to the GraphQL built in scalars and introspection types. All
    # functions below are inline so that this type def cache is within the scope
    # of the closure.

    type_def_cache = {
        'String': GraphQLString,
        'Int': GraphQLInt,
        'Float': GraphQLFloat,
        'Boolean': GraphQLBoolean,
        'ID': GraphQLID,
        '__Schema': __Schema,
        '__Directive': __Directive,
        '__DirectiveLocation': __DirectiveLocation,
        '__Type': __Type,
        '__Field': __Field,
        '__InputValue': __InputValue,
        '__EnumValue': __EnumValue,
        '__TypeKind': __TypeKind,
    }

    # Get the root Query, Mutation, and Subscription types.
    query_type = get_type_from_def(schema.get_query_type())

    existing_mutation_type = schema.get_mutation_type()
    mutationType = existing_mutation_type and get_type_from_def(
        existing_mutation_type) or None

    existing_subscription_type = schema.get_subscription_type()
    subscription_type = existing_subscription_type and get_type_from_def(
        existing_subscription_type) or None

    # Iterate through all types, getting the type definition for each, ensuring
    # that any type not directly referenced by a field will get created.
    types = [get_type_from_def(_def) for _def in schema.get_type_map().values()]

    # Do the same with new types, appending to the list of defined types.
    types += [get_type_from_AST(_def) for _def in type_definition_map.values()]

    # Then produce and return a Schema with these types.
    return GraphQLSchema(
        query=query_type,
        mutation=mutationType,
        subscription=subscription_type,
        # Copy directives.
        directives=schema.get_directives(),
        types=types
    )


def cannot_execute_client_schema(*args, **kwargs):
    raise Exception('Client Schema cannot be used for execution.')



"""
normalize.
"""

import ast
from functools import wraps
import sys
from typing import Callable, TypeVar

import requests
from wandb import env
from wandb.errors import CommError
from wandb_gql.client import RetryError

_F = TypeVar("_F", bound=Callable)


def normalize_exceptions(func: _F) -> _F:
    """Function decorator for catching common errors and re-raising as wandb.Error"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        message = "Whoa, you found a bug."
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as err:
            raise CommError(err.response, err)
        except RetryError as err:
            if (
                "response" in dir(err.last_exception)
                and err.last_exception.response is not None
            ):
                try:
                    message = err.last_exception.response.json().get(
                        "errors", [{"message": message}]
                    )[0]["message"]
                except ValueError:
                    message = err.last_exception.response.text
            else:
                message = err.last_exception

            if env.is_debug():
                raise err.last_exception.with_traceback(sys.exc_info()[2])
            else:
                raise CommError(message, err.last_exception).with_traceback(
                    sys.exc_info()[2]
                )
        except Exception as err:
            # gql raises server errors with dict's as strings...
            if len(err.args) > 0:
                payload = err.args[0]
            else:
                payload = err
            if str(payload).startswith("{"):
                message = ast.literal_eval(str(payload))["message"]
            else:
                message = str(err)
            if env.is_debug():
                raise
            else:
                raise CommError(message, err).with_traceback(sys.exc_info()[2])

    return wrapper



import logging
import os
import pprint
import queue
import socket
import threading
import time
from typing import Any, Optional

import wandb
from wandb import wandb_lib  # type: ignore
from wandb.errors import SweepError
from wandb.wandb_agent import Agent as LegacySweepAgent

from .scheduler import Scheduler, SchedulerState, SimpleRunState, SweepRun

logger = logging.getLogger(__name__)


class SweepScheduler(Scheduler):
    """A SweepScheduler is a controller/agent that will populate a Launch RunQueue with
    launch jobs it creates from run suggestions it pulls from an internal sweeps RunQueue.
    """

    def __init__(
        self,
        *args: Any,
        sweep_id: Optional[str] = None,
        heartbeat_thread_sleep: int = 3,
        heartbeat_queue_timeout: int = 3,
        main_thread_sleep: int = 3,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        # Make sure the provided sweep_id corresponds to a valid sweep
        found = self._api.sweep(
            sweep_id, "{}", entity=self._entity, project=self._project
        )
        if not found:
            raise SweepError(
                f"Could not find sweep {self._entity}/{self._project}/{sweep_id}"
            )
        self._sweep_id = sweep_id
        self._heartbeat_thread_sleep: int = heartbeat_thread_sleep
        self._heartbeat_queue_timeout: int = heartbeat_queue_timeout
        self._main_thread_sleep: int = main_thread_sleep

    def _start(self) -> None:
        self._heartbeat_agent = self._api.register_agent(
            socket.gethostname(),  # host
            sweep_id=self._sweep_id,
            project_name=self._project,
            entity=self._entity,
        )
        self._heartbeat_agent_id = self._heartbeat_agent["id"]
        # Thread will pop items off the Sweeps RunQueue using AgentHeartbeat
        # and put them in this internal queue, which will be used to populate
        # the Launch RunQueue
        self._heartbeat_queue: "queue.Queue[SweepRun]" = queue.Queue()
        self._heartbeat_thread = threading.Thread(target=self._heartbeat)
        self._heartbeat_thread.daemon = True
        self._heartbeat_thread.start()

    def _heartbeat(self) -> None:
        while True:
            if not self.is_alive():
                return
            # AgentHeartbeat wants dict of runs which are running or queued
            _run_states = {}
            for run_id, run in self._runs.items():
                if run.state == SimpleRunState.ALIVE:
                    _run_states[run_id] = True
            _msg = f"AgentHeartbeat sending: \n{pprint.pformat(_run_states)}\n"
            logger.debug(_msg)
            commands = self._api.agent_heartbeat(
                self._heartbeat_agent_id, {}, _run_states
            )
            if commands:
                _msg = f"AgentHeartbeat received {len(commands)} commands: \n{pprint.pformat(commands)}\n"
                logger.debug(_msg)
                for command in commands:
                    _type = command.get("type")
                    # type can be one of "run", "resume", "stop", "exit"
                    if _type == "exit":
                        self.state = SchedulerState.COMPLETED
                        self.exit()
                        return
                    run = SweepRun(
                        id=command.get("run_id"),
                        args=command.get("args"),
                        logs=command.get("logs"),
                        program=command.get("program"),
                    )
                    self._runs[run.id] = run
                    if _type in ["run", "resume"]:
                        self._heartbeat_queue.put(run)
                    elif _type == "stop":
                        self._stop_run(run.id)
                        continue
            time.sleep(self._heartbeat_thread_sleep)

    def _run(self) -> None:
        try:
            run: SweepRun = self._heartbeat_queue.get(
                timeout=self._heartbeat_queue_timeout
            )
        except queue.Empty:
            _msg = "No jobs in Sweeps RunQueue, waiting..."
            logger.debug(_msg)
            wandb.termlog(_msg)
            time.sleep(self._main_thread_sleep)
            return
        # If run is already stopped just ignore the request
        if run.state in [
            SimpleRunState.DEAD,
            SimpleRunState.UNKNOWN,
        ]:
            return
        _msg = f"Converting Sweep Run (RunID:{run.id}) to Launch Job"
        logger.debug(_msg)
        wandb.termlog(_msg)
        # This is actually what populates the wandb config
        # since it is used in wandb.init()
        sweep_param_path = os.path.join(
            os.environ.get(wandb.env.DIR, os.getcwd()),
            "wandb",
            f"sweep-{self._sweep_id}",
            f"config-{run.id}.yaml",
        )
        wandb.termlog(f"Saving params to {sweep_param_path}")
        wandb_lib.config_util.save_config_file_from_dict(sweep_param_path, run.args)
        # Construct entry point using legacy sweeps utilities
        command_args = LegacySweepAgent._create_command_args({"args": run.args})["args"]
        entry_point = ["python", run.program] + command_args
        _ = self._add_to_launch_queue(
            uri=os.environ.get(wandb.env.DIR, os.getcwd()),
            resource="local-process",
            entry_point=entry_point,
            run_id=run.id,
            params=run.args,
        )

    def _exit(self) -> None:
        self.state = SchedulerState.COMPLETED
